from functools import partial

from .model import build_model as _capacity_expansion
from .distributed import build_modular_model as build_distributed_model
from .quarter_distributed import build_modular_model as build_quarter_distributed_model

build_cap_expand_growth = partial(_capacity_expansion, case="Growth")
build_cap_expand_dip = partial(_capacity_expansion, case="Dip")
build_cap_expand_decay = partial(_capacity_expansion, case="Decay")

__all__ = ['build_cap_expand_growth', 'build_cap_expand_dip', 'build_cap_expand_decay', 'build_distributed_model',
           'build_quarter_distributed_model']



from pyomo.environ import *
from pyomo.gdp import *
from pyomo.common.fileutils import this_file_dir
from os.path import join

# Medium-term Purchasing Contracts problem from http://minlp.org/library/lib.php?lib=GDP
# This model maximizes profit in a short-term horizon in which various contracts
# are available for purchasing raw materials. The model decides inventory levels,
# amounts to purchase, amount sold, and flows through the process nodes while
# maximizing profit. The four different contracts available are:
# FIXED PRICE CONTRACT: buy as much as you want at constant price
# DISCOUNT CONTRACT: quantities below minimum amount cost RegPrice. Any additional quantity
# above min amount costs DiscoutPrice.
# BULK CONTRACT: If more than min amount is purchased, whole purchase is at discount price.
# FIXED DURATION CONTRACT: Depending on length of time contract is valid, there is a purchase
# price during that time and min quantity that must be purchased


# This version of the model is a literal transcription of what is in
# ShortTermContractCH.gms from the website. Some data is hardcoded into this model,
# most notably the process structure itself and the mass balance information.

def build_model():
    model = AbstractModel()

    # Constants (data that was hard-coded in GAMS model)
    AMOUNT_UB = 1000
    COST_UB = 1e4
    MAX_AMOUNT_FP = 1000
    MIN_AMOUNT_FD_1MONTH = 0

    RandomConst_Line264 = 0.17
    RandomConst_Line265 = 0.83

    ###################
    # Sets
    ###################

    # T
    # t in GAMS
    model.TimePeriods = Set(ordered=True)

    # Available length contracts
    # p in GAMS
    model.Contracts_Length = Set()

    # JP
    # final(j) in GAMS
    # Finished products
    model.Products = Set()

    # JM
    # rawmat(J) in GAMS
    # Set of Raw Materials-- raw materials, intermediate products, and final products partition J
    model.RawMaterials = Set()

    # C
    # c in GAMS
    model.Contracts = Set()

    # I
    # i in GAMS
    model.Processes = Set()

    # J
    # j in GAMS
    model.Streams = Set()


    ##################
    # Parameters
    ##################

    # Q_it
    # excap(i) in GAMS
    model.Capacity = Param(model.Processes)

    # u_ijt
    # cov(i) in GAMS
    model.ProcessConstants = Param(model.Processes)

    # a_jt^U and d_jt^U
    # spdm(j,t) in GAMS
    model.SupplyAndDemandUBs = Param(model.Streams, model.TimePeriods, default=0)

    # d_jt^L
    # lbdm(j, t) in GAMS
    model.DemandLB = Param(model.Streams, model.TimePeriods, default=0)

    # delta_it
    # delta(i, t) in GAMS
    # operating cost of process i at time t
    model.OperatingCosts = Param(model.Processes, model.TimePeriods)

    # prices of raw materials under FP contract and selling prices of products
    # pf(j, t) in GAMS
    # omega_jt and pf_jt
    model.Prices = Param(model.Streams, model.TimePeriods, default=0)

    # Price for quantities less than min amount under discount contract
    # pd1(j, t) in GAMS
    model.RegPrice_Discount = Param(model.Streams, model.TimePeriods)
    # Discounted price for the quantity purchased exceeding the min amount
    # pd2(j,t0 in GAMS
    model.DiscountPrice_Discount = Param(model.Streams, model.TimePeriods)

    # Price for quantities below min amount
    # pb1(j,t) in GAMS
    model.RegPrice_Bulk = Param(model.Streams, model.TimePeriods)
    # Price for quantities aboce min amount
    # pb2(j, t) in GAMS
    model.DiscountPrice_Bulk = Param(model.Streams, model.TimePeriods)

    # prices with length contract
    # pl(j, p, t) in GAMS
    model.Prices_Length = Param(model.Streams, model.Contracts_Length, model.TimePeriods, default=0)

    # sigmad_jt
    # sigmad(j, t) in GAMS
    # Minimum quantity of chemical j that must be bought before recieving a Discount under discount contract
    model.MinAmount_Discount = Param(model.Streams, model.TimePeriods, default=0)

    # min quantity to recieve discount under bulk contract
    # sigmab(j, t) in GAMS
    model.MinAmount_Bulk = Param(model.Streams, model.TimePeriods, default=0)

    # min quantity to recieve discount under length contract
    # sigmal(j, p) in GAMS
    model.MinAmount_Length = Param(model.Streams, model.Contracts_Length, default=0)

    # main products of process i
    # These are 1 (true) if stream j is the main product of process i, false otherwise.
    # jm(j, i) in GAMS
    model.MainProducts = Param(model.Streams, model.Processes, default=0)

    # theta_jt
    # psf(j, t) in GAMS
    # Shortfall penalty of product j at time t
    model.ShortfallPenalty = Param(model.Products, model.TimePeriods)

    # shortfall upper bound
    # sfub(j, t) in GAMS
    model.ShortfallUB = Param(model.Products, model.TimePeriods, default=0)

    # epsilon_jt
    # cinv(j, t) in GAMS
    # inventory cost of material j at time t
    model.InventoryCost = Param(model.Streams, model.TimePeriods)

    # invub(j, t) in GAMS
    # inventory upper bound
    model.InventoryLevelUB = Param(model.Streams, model.TimePeriods, default=0)

    ## UPPER BOUNDS HARDCODED INTO GAMS MODEL

    # All of these upper bounds are hardcoded. So I am just leaving them that way.
    # This means they all have to be the same as each other right now.
    def getAmountUBs(model, j, t):
        return AMOUNT_UB

    def getCostUBs(model, j, t):
        return COST_UB

    model.AmountPurchasedUB_FP = Param(model.Streams, model.TimePeriods,
        initialize=getAmountUBs)
    model.AmountPurchasedUB_Discount = Param(model.Streams, model.TimePeriods,
        initialize=getAmountUBs)
    model.AmountPurchasedBelowMinUB_Discount = Param(model.Streams, model.TimePeriods,
        initialize=getAmountUBs)
    model.AmountPurchasedAboveMinUB_Discount = Param(model.Streams, model.TimePeriods,
        initialize=getAmountUBs)
    model.AmountPurchasedUB_FD = Param(model.Streams, model.TimePeriods,
        initialize=getAmountUBs)
    model.AmountPurchasedUB_Bulk = Param(model.Streams, model.TimePeriods,
        initialize=getAmountUBs)

    model.CostUB_FP = Param(model.Streams, model.TimePeriods, initialize=getCostUBs)
    model.CostUB_FD = Param(model.Streams, model.TimePeriods, initialize=getCostUBs)
    model.CostUB_Discount = Param(model.Streams, model.TimePeriods, initialize=getCostUBs)
    model.CostUB_Bulk = Param(model.Streams, model.TimePeriods, initialize=getCostUBs)


    ####################
    #VARIABLES
    ####################

    # prof in GAMS
    # will be objective
    model.Profit = Var()

    # f(j, t) in GAMS
    # mass flow rates in tons per time interval t
    model.FlowRate = Var(model.Streams, model.TimePeriods, within=NonNegativeReals)

    # V_jt
    # inv(j, t) in GAMS
    # inventory level of chemical j at time period t
    def getInventoryBounds(model, i, j):
        return (0, model.InventoryLevelUB[i,j])
    model.InventoryLevel = Var(model.Streams, model.TimePeriods,
        bounds=getInventoryBounds)

    # SF_jt
    # sf(j, t) in GAMS
    # Shortfall of demand for chemical j at time period t
    def getShortfallBounds(model, i, j):
        return(0, model.ShortfallUB[i,j])
    model.Shortfall = Var(model.Products, model.TimePeriods,
        bounds=getShortfallBounds)


    # amounts purchased under different contracts

    # spf(j, t) in GAMS
    # Amount of raw material j bought under fixed price contract at time period t
    def get_FP_bounds(model, j, t):
        return (0, model.AmountPurchasedUB_FP[j,t])
    model.AmountPurchased_FP = Var(model.Streams, model.TimePeriods,
        bounds=get_FP_bounds)

    # spd(j, t) in GAMS
    def get_Discount_Total_bounds(model, j, t):
        return (0, model.AmountPurchasedUB_Discount[j,t])
    model.AmountPurchasedTotal_Discount = Var(model.Streams, model.TimePeriods,
        bounds=get_Discount_Total_bounds)

    # Amount purchased below min amount for discount under discount contract
    # spd1(j, t) in GAMS
    def get_Discount_BelowMin_bounds(model, j, t):
        return (0, model.AmountPurchasedBelowMinUB_Discount[j,t])
    model.AmountPurchasedBelowMin_Discount = Var(model.Streams,
        model.TimePeriods, bounds=get_Discount_BelowMin_bounds)

    # spd2(j, t) in GAMS
    # Amount purchased above min amount for discount under discount contract
    def get_Discount_AboveMin_bounds(model, j, t):
        return (0, model.AmountPurchasedBelowMinUB_Discount[j,t])
    model.AmountPurchasedAboveMin_Discount = Var(model.Streams,
        model.TimePeriods, bounds=get_Discount_AboveMin_bounds)

    # Amount purchased under bulk contract
    # spb(j, t) in GAMS
    def get_bulk_bounds(model, j, t):
        return (0, model.AmountPurchasedUB_Bulk[j,t])
    model.AmountPurchased_Bulk = Var(model.Streams, model.TimePeriods,
        bounds=get_bulk_bounds)

    # spl(j, t) in GAMS
    # Amount purchased under Fixed Duration contract
    def get_FD_bounds(model, j, t):
        return (0, model.AmountPurchasedUB_FD[j,t])
    model.AmountPurchased_FD = Var(model.Streams, model.TimePeriods,
        bounds=get_FD_bounds)


    # costs

    # costpl(j, t) in GAMS
    # cost of variable length contract
    def get_CostUBs_FD(model, j, t):
        return (0, model.CostUB_FD[j,t])
    model.Cost_FD = Var(model.Streams, model.TimePeriods, bounds=get_CostUBs_FD)

    # costpf(j, t) in GAMS
    # cost of fixed duration contract
    def get_CostUBs_FP(model, j, t):
        return (0, model.CostUB_FP[j,t])
    model.Cost_FP = Var(model.Streams, model.TimePeriods, bounds=get_CostUBs_FP)

    # costpd(j, t) in GAMS
    # cost of discount contract
    def get_CostUBs_Discount(model, j, t):
        return (0, model.CostUB_Discount[j,t])
    model.Cost_Discount = Var(model.Streams, model.TimePeriods,
        bounds=get_CostUBs_Discount)

    # costpb(j, t) in GAMS
    # cost of bulk contract
    def get_CostUBs_Bulk(model, j, t):
        return (0, model.CostUB_Bulk[j,t])
    model.Cost_Bulk = Var(model.Streams, model.TimePeriods, bounds=get_CostUBs_Bulk)


    # binary variables

    model.BuyFPContract = RangeSet(0,1)
    model.BuyDiscountContract = Set(initialize=('BelowMin', 'AboveMin', 'NotSelected'))
    model.BuyBulkContract = Set(initialize=('BelowMin', 'AboveMin', 'NotSelected'))
    model.BuyFDContract = Set(initialize=('1Month', '2Month', '3Month', 'NotSelected'))


    ################
    # CONSTRAINTS
    ################

    # Objective: maximize profit
    def profit_rule(model):
        salesIncome = sum(model.Prices[j,t] * model.FlowRate[j,t]
            for j in model.Products for t in model.TimePeriods)
        purchaseCost = sum(model.Cost_FD[j,t]
                for j in model.RawMaterials for t in model.TimePeriods) + \
            sum(model.Cost_Discount[j,t]
                for j in model.RawMaterials for t in model.TimePeriods) + \
            sum(model.Cost_Bulk[j,t]
                for j in model.RawMaterials for t in model.TimePeriods) + \
            sum(model.Cost_FP[j,t]
                for j in model.RawMaterials for t in model.TimePeriods)
        productionCost = sum(model.OperatingCosts[i,t] * sum(model.FlowRate[j,t]
            for j in model.Streams if model.MainProducts[j,i])
            for i in model.Processes for t in model.TimePeriods)
        shortfallCost = sum(model.Shortfall[j,t] * model.ShortfallPenalty[j, t]
            for j in model.Products for t in model.TimePeriods)
        inventoryCost = sum(model.InventoryCost[j,t] * model.InventoryLevel[j,t]
            for j in model.Products for t in model.TimePeriods)
        return salesIncome - purchaseCost - productionCost - inventoryCost - shortfallCost
    model.profit = Objective(rule=profit_rule, sense=maximize)

    # flow of raw materials is the total amount purchased (accross all contracts)
    def raw_material_flow_rule(model, j, t):
        return model.FlowRate[j,t] == model.AmountPurchased_FD[j,t] + \
            model.AmountPurchased_FP[j,t] + model.AmountPurchased_Bulk[j,t] + \
            model.AmountPurchasedTotal_Discount[j,t]
    model.raw_material_flow = Constraint(model.RawMaterials, model.TimePeriods,
        rule=raw_material_flow_rule)

    def discount_amount_total_rule(model, j, t):
        return model.AmountPurchasedTotal_Discount[j,t] == \
            model.AmountPurchasedBelowMin_Discount[j,t] + \
            model.AmountPurchasedAboveMin_Discount[j,t]
    model.discount_amount_total_rule = Constraint(model.RawMaterials, model.TimePeriods,
        rule=discount_amount_total_rule)

    # mass balance equations for each node
    # these are specific to the process network in this example.
    def mass_balance_rule1(model, t):
        return model.FlowRate[1, t] == model.FlowRate[2, t] + model.FlowRate[3, t]
    model.mass_balance1 = Constraint(model.TimePeriods, rule=mass_balance_rule1)

    def mass_balance_rule2(model, t):
        return model.FlowRate[5, t] == model.FlowRate[4, t] + model.FlowRate[8,t]
    model.mass_balance2 = Constraint(model.TimePeriods, rule=mass_balance_rule2)

    def mass_balance_rule3(model, t):
        return model.FlowRate[6, t] == model.FlowRate[7, t]
    model.mass_balance3 = Constraint(model.TimePeriods, rule=mass_balance_rule3)

    def mass_balance_rule4(model, t):
        return model.FlowRate[3, t] == 10*model.FlowRate[5, t]
    model.mass_balance4 = Constraint(model.TimePeriods, rule=mass_balance_rule4)

    # process input/output constraints
    # these are also totally specific to the process network
    def process_balance_rule1(model, t):
        return model.FlowRate[9, t] == model.ProcessConstants[1] * model.FlowRate[2, t]
    model.process_balance1 = Constraint(model.TimePeriods, rule=process_balance_rule1)

    def process_balance_rule2(model, t):
        return model.FlowRate[10, t] == model.ProcessConstants[2] * \
            (model.FlowRate[5, t] + model.FlowRate[3, t])
    model.process_balance2 = Constraint(model.TimePeriods, rule=process_balance_rule2)

    def process_balance_rule3(model, t):
        return model.FlowRate[8, t] == RandomConst_Line264 * \
            model.ProcessConstants[3] * model.FlowRate[7, t]
    model.process_balance3 = Constraint(model.TimePeriods, rule=process_balance_rule3)

    def process_balance_rule4(model, t):
        return model.FlowRate[11, t] == RandomConst_Line265 * \
            model.ProcessConstants[3] * model.FlowRate[7, t]
    model.process_balance4 = Constraint(model.TimePeriods, rule=process_balance_rule4)

    # process capacity contraints
    # these are hardcoded based on the three processes and the process flow structure
    def process_capacity_rule1(model, t):
        return model.FlowRate[9, t] <= model.Capacity[1]
    model.process_capacity1 = Constraint(model.TimePeriods, rule=process_capacity_rule1)

    def process_capacity_rule2(model, t):
        return model.FlowRate[10, t] <= model.Capacity[2]
    model.process_capacity2 = Constraint(model.TimePeriods, rule=process_capacity_rule2)

    def process_capacity_rule3(model, t):
        return model.FlowRate[11, t] + model.FlowRate[8, t] <= model.Capacity[3]
    model.process_capacity3 = Constraint(model.TimePeriods, rule=process_capacity_rule3)

    # Inventory balance of final products
    # again, these are hardcoded.

    def inventory_balance1(model, t):
        prev = 0 if t == min(model.TimePeriods) else model.InventoryLevel[12, t-1]
        return prev + model.FlowRate[9, t] == model.FlowRate[12, t] + model.InventoryLevel[12,t]
    model.inventory_balance1 = Constraint(model.TimePeriods, rule=inventory_balance1)

    def inventory_balance_rule2(model, t):
        if t != 1:
            return Constraint.Skip
        return model.FlowRate[10, t] + model.FlowRate[11, t] == \
            model.InventoryLevel[13,t] + model.FlowRate[13, t]
    model.inventory_balance2 = Constraint(model.TimePeriods, rule=inventory_balance_rule2)

    def inventory_balance_rule3(model, t):
        if t <= 1:
            return Constraint.Skip
        return model.InventoryLevel[13, t-1] + model.FlowRate[10, t] + \
            model.FlowRate[11,t] == model.InventoryLevel[13, t] + model.FlowRate[13, t]
    model.inventory_balance3 = Constraint(model.TimePeriods, rule=inventory_balance_rule3)

    # Max capacities of inventories
    def inventory_capacity_rule(model, j, t):
        return model.InventoryLevel[j,t] <= model.InventoryLevelUB[j,t]
    model.inventory_capacity_rule = Constraint(model.Products, model.TimePeriods, rule=inventory_capacity_rule)

    # Shortfall calculation
    def shortfall_rule(model, j, t):
        return model.Shortfall[j, t] == model.SupplyAndDemandUBs[j, t] - model.FlowRate[j,t]
    model.shortfall = Constraint(model.Products, model.TimePeriods, rule=shortfall_rule)

    # maximum shortfall allowed
    def shortfall_max_rule(model, j, t):
        return model.Shortfall[j, t] <= model.ShortfallUB[j, t]
    model.shortfall_max = Constraint(model.Products, model.TimePeriods, rule=shortfall_max_rule)

    # maxiumum capacities of suppliers
    def supplier_capacity_rule(model, j, t):
        return model.FlowRate[j, t] <= model.SupplyAndDemandUBs[j, t]
    model.supplier_capacity = Constraint(model.RawMaterials, model.TimePeriods, rule=supplier_capacity_rule)

    # demand upper bound
    def demand_UB_rule(model, j, t):
        return model.FlowRate[j, t] <= model.SupplyAndDemandUBs[j,t]
    model.demand_UB = Constraint(model.Products, model.TimePeriods, rule=demand_UB_rule)
    # demand lower bound
    def demand_LB_rule(model, j, t):
        return model.FlowRate[j, t] >= model.DemandLB[j,t]
    model.demand_LB = Constraint(model.Products, model.TimePeriods, rule=demand_LB_rule)


    # FIXED PRICE CONTRACT

    # Disjunction for Fixed Price contract buying options
    def FP_contract_disjunct_rule(disjunct, j, t, buy):
        model = disjunct.model()
        if buy:
            disjunct.c = Constraint(expr=model.AmountPurchased_FP[j,t] <= MAX_AMOUNT_FP)
        else:
            disjunct.c = Constraint(expr=model.AmountPurchased_FP[j,t] == 0)
    model.FP_contract_disjunct = Disjunct(model.RawMaterials, model.TimePeriods,
        model.BuyFPContract, rule=FP_contract_disjunct_rule)

    # Fixed price disjunction
    def FP_contract_rule(model, j, t):
        return [model.FP_contract_disjunct[j,t,buy] for buy in model.BuyFPContract]
    model.FP_disjunction = Disjunction(model.RawMaterials, model.TimePeriods,
        rule=FP_contract_rule)

    # cost constraint for fixed price contract (independent contraint)
    def FP_contract_cost_rule(model, j, t):
        return model.Cost_FP[j,t] == model.AmountPurchased_FP[j,t] * \
            model.Prices[j,t]
    model.FP_contract_cost = Constraint(model.RawMaterials, model.TimePeriods,
        rule=FP_contract_cost_rule)


    # DISCOUNT CONTRACT

    # Disjunction for Discount contract
    def discount_contract_disjunct_rule(disjunct, j, t, buy):
        model = disjunct.model()
        if buy == 'BelowMin':
            disjunct.belowMin = Constraint(
                expr=model.AmountPurchasedBelowMin_Discount[j,t] <= \
                model.MinAmount_Discount[j,t])
            disjunct.aboveMin = Constraint(
                expr=model.AmountPurchasedAboveMin_Discount[j,t] == 0)
        elif buy == 'AboveMin':
            disjunct.belowMin = Constraint(
                expr=model.AmountPurchasedBelowMin_Discount[j,t] == \
                model.MinAmount_Discount[j,t])
            disjunct.aboveMin = Constraint(
                expr=model.AmountPurchasedAboveMin_Discount[j,t] >= 0)
        elif buy == 'NotSelected':
            disjunct.belowMin = Constraint(
                expr=model.AmountPurchasedBelowMin_Discount[j,t] == 0)
            disjunct.aboveMin = Constraint(
                expr=model.AmountPurchasedAboveMin_Discount[j,t] == 0)
        else:
            raise RuntimeError("Unrecognized choice for discount contract: %s" % buy)
    model.discount_contract_disjunct = Disjunct(model.RawMaterials, model.TimePeriods,
        model.BuyDiscountContract, rule=discount_contract_disjunct_rule)

    # Discount contract disjunction
    def discount_contract_rule(model, j, t):
        return [model.discount_contract_disjunct[j,t,buy] \
            for buy in model.BuyDiscountContract]
    model.discount_contract = Disjunction(model.RawMaterials, model.TimePeriods,
        rule=discount_contract_rule)

    # cost constraint for discount contract (independent constraint)
    def discount_cost_rule(model, j, t):
        return model.Cost_Discount[j,t] == model.RegPrice_Discount[j,t] * \
            model.AmountPurchasedBelowMin_Discount[j,t] + \
            model.DiscountPrice_Discount[j,t] * model.AmountPurchasedAboveMin_Discount[j,t]
    model.discount_cost = Constraint(model.RawMaterials, model.TimePeriods,
        rule=discount_cost_rule)


    # BULK CONTRACT

    # Bulk contract buying options disjunct
    def bulk_contract_disjunct_rule(disjunct, j, t, buy):
        model = disjunct.model()
        if buy == 'BelowMin':
            disjunct.amount = Constraint(
                expr=model.AmountPurchased_Bulk[j,t] <= model.MinAmount_Bulk[j,t])
            disjunct.price = Constraint(
                expr=model.Cost_Bulk[j,t] == model.RegPrice_Bulk[j,t] * \
                model.AmountPurchased_Bulk[j,t])
        elif buy == 'AboveMin':
            disjunct.amount = Constraint(
                expr=model.AmountPurchased_Bulk[j,t] >= model.MinAmount_Bulk[j,t])
            disjunct.price = Constraint(
                expr=model.Cost_Bulk[j,t] == model.DiscountPrice_Bulk[j,t] * \
                model.AmountPurchased_Bulk[j,t])
        elif buy == 'NotSelected':
            disjunct.amount = Constraint(expr=model.AmountPurchased_Bulk[j,t] == 0)
            disjunct.price = Constraint(expr=model.Cost_Bulk[j,t] == 0)
        else:
            raise RuntimeError("Unrecognized choice for bulk contract: %s" % buy)
    model.bulk_contract_disjunct = Disjunct(model.RawMaterials, model.TimePeriods,
        model.BuyBulkContract, rule=bulk_contract_disjunct_rule)

    # Bulk contract disjunction
    def bulk_contract_rule(model, j, t):
        return [model.bulk_contract_disjunct[j,t,buy] for buy in model.BuyBulkContract]
    model.bulk_contract = Disjunction(model.RawMaterials, model.TimePeriods,
        rule=bulk_contract_rule)


    # FIXED DURATION CONTRACT

    def FD_1mo_contract(disjunct, j, t):
       model = disjunct.model()
       disjunct.amount1 = Constraint(expr=model.AmountPurchased_FD[j,t] >= \
           MIN_AMOUNT_FD_1MONTH)
       disjunct.price1 = Constraint(expr=model.Cost_FD[j,t] == \
           model.Prices_Length[j,1,t] * model.AmountPurchased_FD[j,t])
    model.FD_1mo_contract = Disjunct(
       model.RawMaterials, model.TimePeriods, rule=FD_1mo_contract)

    def FD_2mo_contract(disjunct, j, t):
       model = disjunct.model()
       disjunct.amount1 = Constraint(expr=model.AmountPurchased_FD[j,t] >= \
           model.MinAmount_Length[j,2])
       disjunct.price1 = Constraint(expr=model.Cost_FD[j,t] == \
           model.Prices_Length[j,2,t] * model.AmountPurchased_FD[j,t])
       # only enforce these if we aren't in the last time period
       if t < model.TimePeriods[-1]:
           disjunct.amount2 = Constraint(expr=model.AmountPurchased_FD[j, t+1] >= \
               model.MinAmount_Length[j,2])
           disjunct.price2 = Constraint(expr=model.Cost_FD[j,t+1] == \
               model.Prices_Length[j,2,t] * model.AmountPurchased_FD[j, t+1])
    model.FD_2mo_contract = Disjunct(
       model.RawMaterials, model.TimePeriods, rule=FD_2mo_contract)

    def FD_3mo_contract(disjunct, j, t):
       model = disjunct.model()
       # NOTE: I think there is a mistake in the GAMS file in line 327.
       # they use the bulk minamount rather than the length one.
       #I am doing the same here for validation purposes.
       disjunct.amount1 = Constraint(expr=model.AmountPurchased_FD[j,t] >= \
           model.MinAmount_Bulk[j,3])
       disjunct.cost1 = Constraint(expr=model.Cost_FD[j,t] == \
           model.Prices_Length[j,3,t] * model.AmountPurchased_FD[j,t])
       # check we aren't in one of the last two time periods
       if t < model.TimePeriods[-1]:
           disjunct.amount2 = Constraint(expr=model.AmountPurchased_FD[j,t+1] >= \
               model.MinAmount_Length[j,3])
           disjunct.cost2 = Constraint(expr=model.Cost_FD[j,t+1] == \
               model.Prices_Length[j,3,t] * model.AmountPurchased_FD[j,t+1])
       if t < model.TimePeriods[-2]:
           disjunct.amount3 = Constraint(expr=model.AmountPurchased_FD[j,t+2] >= \
               model.MinAmount_Length[j,3])
           disjunct.cost3 = Constraint(expr=model.Cost_FD[j,t+2] == \
               model.Prices_Length[j,3,t] * model.AmountPurchased_FD[j,t+2])
    model.FD_3mo_contract = Disjunct(
       model.RawMaterials, model.TimePeriods, rule=FD_3mo_contract)

    def FD_no_contract(disjunct, j, t):
        model = disjunct.model()
        disjunct.amount1 = Constraint(expr=model.AmountPurchased_FD[j,t] == 0)
        disjunct.cost1 = Constraint(expr=model.Cost_FD[j,t] == 0)
        if t < model.TimePeriods[-1]:
            disjunct.amount2 = Constraint(expr=model.AmountPurchased_FD[j,t+1] == 0)
            disjunct.cost2 = Constraint(expr=model.Cost_FD[j,t+1] == 0)
        if t < model.TimePeriods[-2]:
            disjunct.amount3 = Constraint(expr=model.AmountPurchased_FD[j,t+2] == 0)
            disjunct.cost3 = Constraint(expr=model.Cost_FD[j,t+2] == 0)
    model.FD_no_contract = Disjunct(
       model.RawMaterials, model.TimePeriods, rule=FD_no_contract)

    def FD_contract(model, j, t):
       return [ model.FD_1mo_contract[j,t], model.FD_2mo_contract[j,t],
                model.FD_3mo_contract[j,t], model.FD_no_contract[j,t], ]
    model.FD_contract = Disjunction(model.RawMaterials, model.TimePeriods,
       rule=FD_contract)

    return model


def build_concrete():
    return build_model().create_instance(join(this_file_dir(), 'med_term_purchasing.dat'))


if __name__ == "__main__":
    m = build_concrete()
    TransformationFactory('gdp.bigm').apply_to(m)
    SolverFactory('gams').solve(m, solver='baron', tee=True, add_options=['option optcr=1e-6;'])
    m.profit.display()



"""Distillation column model for 2018 PSE conference"""

from __future__ import division

from math import fabs

from pyomo.environ import SolverFactory, Suffix, value
from pyomo.util.infeasible import log_infeasible_constraints

from gdplib.gdp_col.column import build_column
from gdplib.gdp_col.initialize import initialize


def main():
    m = build_column(min_trays=8, max_trays=17, xD=0.95, xB=0.95)
    # Fix feed conditions
    m.feed['benzene'].fix(50)
    m.feed['toluene'].fix(50)
    m.T_feed.fix(368)
    m.feed_vap_frac.fix(0.40395)
    # Initial values of reflux and reboil ratios
    m.reflux_ratio.set_value(1.4)
    m.reboil_ratio.set_value(1.3)
    # Fix to be total condenser
    m.partial_cond.deactivate()
    m.total_cond.indicator_var.fix(1)
    # Give initial values of the tray existence/absence
    for t in m.conditional_trays:
        m.tray[t].indicator_var.set_value(1)
        m.no_tray[t].indicator_var.set_value(0)
    # Run custom initialization routine
    initialize(m)

    m.BigM = Suffix(direction=Suffix.LOCAL)
    m.BigM[None] = 100

    SolverFactory('gdpopt').solve(
        m, tee=True, init_strategy='fix_disjuncts',
        mip_solver='glpk')
    log_infeasible_constraints(m, tol=1E-3)
    display_column(m)
    return m


def display_column(m):
    print('Objective: %s' % value(m.obj))
    print('Qc: {: >3.0f}kW  DB: {: >3.0f} DT: {: >3.0f} dis: {: >3.0f}'
          .format(value(m.Qc * 1E3),
                  value(m.D['benzene']),
                  value(m.D['toluene']),
                  value(m.dis)))
    for t in reversed(list(m.trays)):
        print('T{: >2.0f}-{:1.0g} T: {: >3.0f} '
              'F: {: >4.0f} '
              'L: {: >4.0f} V: {: >4.0f} '
              'xB: {: >3.0f} xT: {: >3.0f} yB: {: >3.0f} yT: {: >3.0f}'
              .format(t,
                      fabs(value(m.tray[t].indicator_var))
                      if t in m.conditional_trays else 1,
                      value(m.T[t]) - 273.15,
                      value(sum(m.feed[c] for c in m.comps))
                      if t == m.feed_tray else 0,
                      value(m.liq[t]),
                      value(m.vap[t]),
                      value(m.x['benzene', t]) * 100,
                      value(m.x['toluene', t]) * 100,
                      value(m.y['benzene', t]) * 100,
                      value(m.y['toluene', t]) * 100
                      ))
    print('Qb: {: >3.0f}kW  BB: {: > 3.0f} BT: {: >3.0f} bot: {: >3.0f}'
          .format(value(m.Qb * 1E3),
                  value(m.B['benzene']),
                  value(m.B['toluene']),
                  value(m.bot)))
    for t in reversed(list(m.trays)):
        print('T{: >2.0f}-{:1.0g} '
              'FB: {: >3.0f} FT: {: >3.0f} '
              'LB: {: >4.0f} LT: {: >4.0f} VB: {: >4.0f} VT: {: >4.0f}'
              .format(t,
                      fabs(value(m.tray[t].indicator_var))
                      if t in m.conditional_trays else 1,
                      value(m.feed['benzene']) if t == m.feed_tray else 0,
                      value(m.feed['toluene']) if t == m.feed_tray else 0,
                      value(m.L['benzene', t]),
                      value(m.L['toluene', t]),
                      value(m.V['benzene', t]),
                      value(m.V['toluene', t])
                      ))
    print('RF: {: >3.2f} RB: {: >3.2f}'
          .format(value(m.reflux_frac / (1 - m.reflux_frac)),
                  value(m.boilup_frac / (1 - m.boilup_frac))))


if __name__ == "__main__":
    m = main()



from __future__ import division

from pyomo.environ import (ConcreteModel, NonNegativeReals, Set, SolverFactory,
                           Var, log, sqrt)


def calculate_Fenske(xD, xB):
    m = ConcreteModel()
    min_T, max_T = 300, 400
    m.comps = Set(initialize=['benzene', 'toluene'])
    m.trays = Set(initialize=['condenser', 'reboiler'])
    m.Kc = Var(
        m.comps, m.trays, doc='Phase equilibrium constant',
        domain=NonNegativeReals, initialize=1, bounds=(0, 1000))
    m.T = Var(m.trays, doc='Temperature [K]',
              domain=NonNegativeReals,
              bounds=(min_T, max_T))
    m.T['condenser'].fix(82 + 273.15)
    m.T['reboiler'].fix(108 + 273.15)
    m.P = Var(doc='Pressure [bar]',
              bounds=(0, 5))
    m.P.fix(1.01)
    m.T_ref = 298.15
    m.gamma = Var(
        m.comps, m.trays,
        doc='liquid activity coefficent of component on tray',
        domain=NonNegativeReals, bounds=(0, 10), initialize=1)
    m.Pvap = Var(
        m.comps, m.trays,
        doc='pure component vapor pressure of component on tray in bar',
        domain=NonNegativeReals, bounds=(1E-3, 5), initialize=0.4)
    m.Pvap_X = Var(
        m.comps, m.trays,
        doc='Related to fraction of critical temperature (1 - T/Tc)',
        bounds=(0.25, 0.5), initialize=0.4)

    m.pvap_const = {
        'benzene': {'A': -6.98273, 'B': 1.33213, 'C': -2.62863,
                    'D': -3.33399, 'Tc': 562.2, 'Pc': 48.9},
        'toluene': {'A': -7.28607, 'B': 1.38091, 'C': -2.83433,
                    'D': -2.79168, 'Tc': 591.8, 'Pc': 41.0}}

    @m.Constraint(m.comps, m.trays)
    def phase_equil_const(_, c, t):
        return m.Kc[c, t] * m.P == (
            m.gamma[c, t] * m.Pvap[c, t])

    @m.Constraint(m.comps, m.trays)
    def Pvap_relation(_, c, t):
        k = m.pvap_const[c]
        x = m.Pvap_X[c, t]
        return (log(m.Pvap[c, t]) - log(k['Pc'])) * (1 - x) == (
            k['A'] * x +
            k['B'] * x ** 1.5 +
            k['C'] * x ** 3 +
            k['D'] * x ** 6)

    @m.Constraint(m.comps, m.trays)
    def Pvap_X_defn(_, c, t):
        k = m.pvap_const[c]
        return m.Pvap_X[c, t] == 1 - m.T[t] / k['Tc']

    @m.Constraint(m.comps, m.trays)
    def gamma_calc(_, c, t):
        return m.gamma[c, t] == 1

    m.relative_volatility = Var(m.trays, domain=NonNegativeReals)

    @m.Constraint(m.trays)
    def relative_volatility_calc(_, t):
        return m.Kc['benzene', t] == (
            m.Kc['toluene', t] * m.relative_volatility[t])

    @m.Expression()
    def fenske(_):
        return log((xD / (1 - xD)) * (xB / (1 - xB))) / (
            log(sqrt(m.relative_volatility['condenser'] *
                     m.relative_volatility['reboiler'])))

    SolverFactory('ipopt').solve(m, tee=True)
    from pyomo.util.infeasible import log_infeasible_constraints
    log_infeasible_constraints(m, tol=1E-3)
    m.fenske.display()


if __name__ == '__main__':
    m = calculate_Fenske(0.95, 0.95)



from __future__ import division

import os

import pandas as pd

from pyomo.environ import (
    ConcreteModel, Constraint, Integers, NonNegativeReals, Objective, Param,
    RangeSet, Set, SolverFactory, Suffix, TransformationFactory, Var, exp, log,
    sqrt, summation, value
)
from .util import alphanum_sorted
from pyomo.environ import TerminationCondition as tc


def build_model():
    m = ConcreteModel()
    m.BigM = Suffix(direction=Suffix.LOCAL)

    m.periods_per_year = Param(initialize=4, doc="Quarters per year")
    m.project_life = Param(initialize=15, doc="Years")
    m.time = RangeSet(0, m.periods_per_year *
                      m.project_life - 1, doc="Time periods")
    m.discount_rate = Param(initialize=0.08, doc="8%")
    m.learning_rate = Param(initialize=0.1, doc="Fraction discount for doubling of quantity")

    m.module_setup_time = Param(
        initialize=1, doc="1 quarter for module transfer")

    @m.Param(m.time)
    def discount_factor(m, t):
        return (1 + m.discount_rate / m.periods_per_year) ** (-t / m.periods_per_year)

    xlsx_data = pd.read_excel(os.path.join(os.path.dirname(__file__), "data.xlsx"), sheet_name=None)
    module_sheet = xlsx_data['modules'].set_index('Type')
    m.module_types = Set(initialize=module_sheet.columns.tolist(),)

    @m.Param(m.module_types)
    def module_base_cost(m, mtype):
        return float(module_sheet[mtype]['Capital Cost [MM$]'])

    @m.Param(m.module_types, doc="Natural gas consumption per module of this type [MMSCF/d]")
    def unit_gas_consumption(m, mtype):
        return float(module_sheet[mtype]['Nat Gas [MMSCF/d]'])

    @m.Param(m.module_types, doc="Gasoline production per module of this type [kBD]")
    def gasoline_production(m, mtype):
        return float(module_sheet[mtype]['Gasoline [kBD]'])

    @m.Param(m.module_types, doc="Overall conversion of natural gas into gasoline per module of this type [kB/MMSCF]")
    def module_conversion(m, mtype):
        return float(module_sheet[mtype]['Conversion [kB/MMSCF]'])

    site_sheet = xlsx_data['sites'].set_index('Potential site')
    m.potential_sites = Set(initialize=site_sheet.index.tolist())
    m.site_pairs = Set(
        doc="Pairs of potential sites",
        initialize=m.potential_sites * m.potential_sites,
        filter=lambda _, x, y: not x == y)

    @m.Param(m.potential_sites)
    def site_x(m, site):
        return float(site_sheet['x'][site])

    @m.Param(m.potential_sites)
    def site_y(m, site):
        return float(site_sheet['y'][site])

    well_sheet = xlsx_data['wells'].set_index('Well')
    m.well_clusters = Set(initialize=well_sheet.index.tolist())

    @m.Param(m.well_clusters)
    def well_x(m, well):
        return float(well_sheet['x'][well])

    @m.Param(m.well_clusters)
    def well_y(m, well):
        return float(well_sheet['y'][well])

    sched_sheet = xlsx_data['well-schedule']
    decay_curve = [1] + [3.69 * exp(-1.31 * (t + 1) ** 0.292) for t in range(m.project_life * 12)]
    well_profiles = {well: [0 for _ in decay_curve] for well in m.well_clusters}
    for _, well_info in sched_sheet.iterrows():
        start_time = int(well_info['Month'])
        prod = [0] * start_time + decay_curve[:len(decay_curve) - start_time]
        prod = [x * float(well_info['max prod [MMSCF/d]']) for x in prod]
        current_profile = well_profiles[well_info['well-cluster']]
        well_profiles[well_info['well-cluster']] = [val + prod[i] for i, val in enumerate(current_profile)]

    @m.Param(m.well_clusters, m.time, doc="Supply of gas from well cluster [MMSCF/day]")
    def gas_supply(m, well, t):
        return sum(well_profiles[well][t * 3:t * 3 + 2]) / 3

    mkt_sheet = xlsx_data['markets'].set_index('Market')
    m.markets = Set(initialize=mkt_sheet.index.tolist())

    @m.Param(m.markets)
    def mkt_x(m, mkt):
        return float(mkt_sheet['x'][mkt])

    @m.Param(m.markets)
    def mkt_y(m, mkt):
        return float(mkt_sheet['y'][mkt])

    @m.Param(m.markets, doc="Gasoline demand [kBD]")
    def mkt_demand(m, mkt):
        return float(mkt_sheet['demand [kBD]'][mkt])

    m.sources = Set(initialize=m.well_clusters | m.potential_sites)
    m.destinations = Set(initialize=m.potential_sites | m.markets)

    @m.Param(m.sources, m.destinations, doc="Distance [mi]")
    def distance(m, src, dest):
        if src in m.well_clusters:
            src_x = m.well_x[src]
            src_y = m.well_y[src]
        else:
            src_x = m.site_x[src]
            src_y = m.site_y[src]
        if dest in m.markets:
            dest_x = m.mkt_x[dest]
            dest_y = m.mkt_y[dest]
        else:
            dest_x = m.site_x[dest]
            dest_y = m.site_y[dest]
        return sqrt((src_x - dest_x) ** 2 + (src_y - dest_y) ** 2)

    m.num_modules = Var(
        m.module_types, m.potential_sites, m.time,
        doc="Number of active modules of each type at a site in a period",
        domain=Integers, bounds=(0, 50), initialize=1)
    m.modules_transferred = Var(
        m.module_types, m.site_pairs, m.time,
        doc="Number of module transfers initiated from one site to another in a period.",
        domain=Integers, bounds=(0, 15), initialize=0)
    m.modules_purchased = Var(
        m.module_types, m.potential_sites, m.time,
        doc="Number of modules of each type purchased for a site in a period",
        domain=Integers, bounds=(0, 30), initialize=1)

    m.pipeline_unit_cost = Param(doc="MM$/mile", initialize=2)

    @m.Param(m.time, doc="Module transport cost per mile [M$/100 miles]")
    def module_transport_distance_cost(m, t):
        return 50 * m.discount_factor[t]

    @m.Param(m.time, doc="Module transport cost per unit [MM$/module]")
    def module_transport_unit_cost(m, t):
        return 3 * m.discount_factor[t]

    @m.Param(m.time, doc="Stranded gas price [$/MSCF]")
    def nat_gas_price(m, t):
        return 5 * m.discount_factor[t]

    @m.Param(m.time, doc="Gasoline price [$/gal]")
    def gasoline_price(m, t):
        return 2.5 * m.discount_factor[t]

    @m.Param(m.time, doc="Gasoline transport cost [$/gal/100 miles]")
    def gasoline_tranport_cost(m, t):
        return 0.045 * m.discount_factor[t]

    m.gal_per_bbl = Param(initialize=42, doc="Gallons per barrel")
    m.days_per_period = Param(initialize=90, doc="Days in a production period")

    m.learning_factor = Var(
        m.module_types,
        doc="Fraction of cost due to economies of mass production",
        domain=NonNegativeReals, bounds=(0, 1), initialize=1)

    @m.Disjunct(m.module_types)
    def mtype_exists(disj, mtype):
        disj.learning_factor_calc = Constraint(
            expr=m.learning_factor[mtype] == (1 - m.learning_rate) ** (
                log(sum(m.modules_purchased[mtype, :, :])) / log(2)))
        m.BigM[disj.learning_factor_calc] = 1
        disj.require_module_purchases = Constraint(
            expr=sum(m.modules_purchased[mtype, :, :]) >= 1)

    @m.Disjunct(m.module_types)
    def mtype_absent(disj, mtype):
        disj.constant_learning_factor = Constraint(
            expr=m.learning_factor[mtype] == 1)

    @m.Disjunction(m.module_types)
    def mtype_existence(m, mtype):
        return [m.mtype_exists[mtype], m.mtype_absent[mtype]]

    @m.Expression(m.module_types, m.time, doc="Module unit cost [MM$/module]")
    def module_unit_cost(m, mtype, t):
        return m.module_base_cost[mtype] * m.learning_factor[mtype] * m.discount_factor[t]

    m.production = Var(
        m.potential_sites, m.time,
        doc="Production of gasoline in a time period [kBD]",
        domain=NonNegativeReals, bounds=(0, 30), initialize=10)
    m.gas_consumption = Var(
        m.potential_sites, m.module_types, m.time,
        doc="Consumption of natural gas by each module type "
        "at each site in a time period [MMSCF/d]",
        domain=NonNegativeReals, bounds=(0, 250), initialize=50)
    m.gas_flows = Var(
        m.well_clusters, m.potential_sites, m.time,
        doc="Flow of gas from a well cluster to a site [MMSCF/d]",
        domain=NonNegativeReals, bounds=(0, 200), initialize=15)
    m.product_flows = Var(
        m.potential_sites, m.markets, m.time,
        doc="Product shipments from a site to a market in a period [kBD]",
        domain=NonNegativeReals, bounds=(0, 30), initialize=10)

    @m.Constraint(m.potential_sites, m.module_types, m.time)
    def consumption_capacity(m, site, mtype, t):
        return m.gas_consumption[site, mtype, t] <= (
            m.num_modules[mtype, site, t] * m.unit_gas_consumption[mtype])

    @m.Constraint(m.potential_sites, m.time)
    def production_limit(m, site, t):
        return m.production[site, t] <= sum(
            m.gas_consumption[site, mtype, t] * m.module_conversion[mtype]
            for mtype in m.module_types)

    @m.Expression(m.potential_sites, m.time)
    def capacity(m, site, t):
        return sum(
            m.num_modules[mtype, site, t] * m.unit_gas_consumption[mtype]
            * m.module_conversion[mtype] for mtype in m.module_types)

    @m.Constraint(m.potential_sites, m.time)
    def gas_supply_meets_consumption(m, site, t):
        return sum(m.gas_consumption[site, :, t]) == sum(m.gas_flows[:, site, t])

    @m.Constraint(m.well_clusters, m.time)
    def gas_supply_limit(m, well, t):
        return sum(m.gas_flows[well, site, t]
                   for site in m.potential_sites) <= m.gas_supply[well, t]

    @m.Constraint(m.potential_sites, m.time)
    def gasoline_production_requirement(m, site, t):
        return sum(m.product_flows[site, mkt, t]
                   for mkt in m.markets) == m.production[site, t]

    @m.Constraint(m.potential_sites, m.module_types, m.time)
    def module_balance(m, site, mtype, t):
        if t >= m.module_setup_time:
            modules_added = m.modules_purchased[
                mtype, site, t - m.module_setup_time]
            modules_transferred_in = sum(
                m.modules_transferred[
                    mtype, from_site, to_site, t - m.module_setup_time]
                for from_site, to_site in m.site_pairs if to_site == site)
        else:
            modules_added = 0
            modules_transferred_in = 0
        if t >= 1:
            existing_modules = m.num_modules[mtype, site, t - 1]
        else:
            existing_modules = 0
        modules_transferred_out = sum(
            m.modules_transferred[mtype, from_site, to_site, t]
            for from_site, to_site in m.site_pairs if from_site == site)

        return m.num_modules[mtype, site, t] == (
            existing_modules + modules_added
            + modules_transferred_in - modules_transferred_out)

    @m.Disjunct(m.potential_sites)
    def site_active(disj, site):
        pass

    @m.Disjunct(m.potential_sites)
    def site_inactive(disj, site):
        disj.no_production = Constraint(
            expr=sum(m.production[site, :]) == 0)
        disj.no_gas_consumption = Constraint(
            expr=sum(m.gas_consumption[site, :, :]) == 0)
        disj.no_gas_flows = Constraint(
            expr=sum(m.gas_flows[:, site, :]) == 0)
        disj.no_product_flows = Constraint(
            expr=sum(m.product_flows[site, :, :]) == 0)
        disj.no_modules = Constraint(
            expr=sum(m.num_modules[:, site, :]) == 0)
        disj.no_modules_transferred = Constraint(
            expr=sum(
                m.modules_transferred[mtypes, from_site, to_site, t]
                for mtypes in m.module_types
                for from_site, to_site in m.site_pairs
                for t in m.time
                if from_site == site or to_site == site) == 0)
        disj.no_modules_purchased = Constraint(
            expr=sum(
                m.modules_purchased[mtype, site, t]
                for mtype in m.module_types for t in m.time) == 0)

    @m.Disjunction(m.potential_sites)
    def site_active_or_not(m, site):
        return [m.site_active[site], m.site_inactive[site]]

    @m.Disjunct(m.well_clusters, m.potential_sites)
    def pipeline_exists(disj, well, site):
        pass

    @m.Disjunct(m.well_clusters, m.potential_sites)
    def pipeline_absent(disj, well, site):
        disj.no_natural_gas_flow = Constraint(
            expr=sum(m.gas_flows[well, site, t] for t in m.time) == 0)

    @m.Disjunction(m.well_clusters, m.potential_sites)
    def pipeline_existence(m, well, site):
        return [m.pipeline_exists[well, site], m.pipeline_absent[well, site]]

    # Objective Function Construnction
    @m.Expression(m.potential_sites, doc="MM$")
    def product_revenue(m, site):
        return sum(
            m.product_flows[site, mkt, t]  # kBD
            * 1000  # bbl/kB
            / 1E6  # $ to MM$
            * m.days_per_period
            * m.gasoline_price[t] * m.gal_per_bbl
            for mkt in m.markets
            for t in m.time)

    @m.Expression(m.potential_sites, doc="MM$")
    def raw_material_cost(m, site):
        return sum(
            m.gas_consumption[site, mtype, t] * m.days_per_period
            / 1E6  # $ to MM$
            * m.nat_gas_price[t]
            * 1000  # MMSCF to MSCF
            for mtype in m.module_types for t in m.time)

    @m.Expression(
        m.potential_sites, m.markets,
        doc="Aggregate cost to transport gasoline from a site to market [MM$]")
    def product_transport_cost(m, site, mkt):
        return sum(
            m.product_flows[site, mkt, t] * m.gal_per_bbl
            * 1000  # bbl/kB
            / 1E6  # $ to MM$
            * m.distance[site, mkt] / 100 * m.gasoline_tranport_cost[t]
            for t in m.time)

    @m.Expression(m.well_clusters, m.potential_sites, doc="MM$")
    def pipeline_construction_cost(m, well, site):
        return (m.pipeline_unit_cost * m.distance[well, site]
                * m.pipeline_exists[well, site].indicator_var)

    # Module transport cost
    @m.Expression(m.site_pairs, doc="MM$")
    def module_relocation_cost(m, from_site, to_site):
        return sum(
            m.modules_transferred[mtype, from_site, to_site, t]
            * m.distance[from_site, to_site] / 100
            * m.module_transport_distance_cost[t]
            / 1E3  # M$ to MM$
            + m.modules_transferred[mtype, from_site, to_site, t]
            * m.module_transport_unit_cost[t]
            for mtype in m.module_types
            for t in m.time)

    @m.Expression(m.potential_sites, doc="MM$")
    def module_purchase_cost(m, site):
        return sum(
            m.module_unit_cost[mtype, t] * m.modules_purchased[mtype, site, t]
            for mtype in m.module_types
            for t in m.time)

    @m.Expression(doc="MM$")
    def profit(m):
        return (
            summation(m.product_revenue)
            - summation(m.raw_material_cost)
            - summation(m.product_transport_cost)
            - summation(m.pipeline_construction_cost)
            - summation(m.module_relocation_cost)
            - summation(m.module_purchase_cost)
        )

    m.neg_profit = Objective(expr=-m.profit)

    # Tightening constraints
    @m.Constraint(doc="Limit total module purchases over project span.")
    def restrict_module_purchases(m):
        return sum(m.modules_purchased[...]) <= 5

    @m.Constraint(m.site_pairs, doc="Limit transfers between any two sites")
    def restrict_module_transfers(m, from_site, to_site):
        return sum(m.modules_transferred[:, from_site, to_site, :]) <= 5

    return m


if __name__ == "__main__":
    m = build_model()

    # Restrict number of module types; A, R, S, U
    # valid_modules = ['A500', 'A1000', 'A2000', 'A5000']
    # valid_modules = ['A500', 'R500', 'A5000', 'R5000']
    # valid_modules = ['U500', 'U5000']
    # valid_modules = ['U100', 'U250']
    # valid_modules = ['U1000']
    # valid_modules = ['U500']
    valid_modules = ['U250']
    # valid_modules = ['U100']
    for mtype in m.module_types - valid_modules:
        m.gas_consumption[:, mtype, :].fix(0)
        m.num_modules[mtype, :, :].fix(0)
        m.modules_transferred[mtype, :, :, :].fix(0)
        m.modules_purchased[mtype, :, :].fix(0)
        m.mtype_exists[mtype].deactivate()
        m.mtype_absent[mtype].indicator_var.fix(1)



import gdplib.mod_hens
import gdplib.modprodnet
import gdplib.biofuel
import gdplib.logical  # Requires logical expression system
import gdplib.pyomo_examples
import gdplib.stranded_gas  # Requires logical expression system
import gdplib.gdp_col



from __future__ import division

import os

import matplotlib.pyplot as plt
import pandas as pd

from pyomo.environ import (ConcreteModel, Constraint, Expression, Integers,
                           Objective, Param, RangeSet, Set, SolverFactory,
                           Suffix, TerminationCondition, TransformationFactory,
                           Var, sqrt, value)


def build_model():
    m = ConcreteModel()
    m.quarters = RangeSet(0, 40, doc="10 years")
    m.discount_rate = Param(initialize=0.08, doc="8%")
    m.conv_setup_time = Param(initialize=4)
    m.markets = RangeSet(5)

    xls_data = pd.read_excel(
        os.path.join(os.path.dirname(__file__), "quarter_multiple_market_size.xlsx"),
        sheet_name=["demand", "locations"], index_col=0)

    @m.Param(m.markets, m.quarters)
    def market_demand(m, mkt, qtr):
        return float(xls_data["demand"]["market%s" % mkt][qtr])

    @m.Param(m.quarters, doc="Cost of transporting unit of product one mile.")
    def transport_cost(m, qtr):
        return 0.2 / 12 * (1 + m.discount_rate / 4) ** (-qtr / 4)

    m.route_fixed_cost = Param(
        initialize=100,
        doc="Cost of establishing a route from a modular site to a market")

    m.conv_x = Var(bounds=(0, 300), doc="x-coordinate of centralized plant.")
    m.conv_y = Var(bounds=(0, 300), doc="y-coordinate of centralized plant.")

    m.conv_size = Var(bounds=(10, 700), initialize=10,
                      doc="Size of conventional plant.")
    m.conv_cost = Var()
    m.conv_base_cost = Param(initialize=1000, doc="Cost for size 60")
    m.conv_exponent = Param(initialize=0.6)

    m.cost_calc = Constraint(
        expr=m.conv_cost == (
            m.conv_base_cost * (m.conv_size / 60) ** m.conv_exponent))

    @m.Param(m.markets)
    def mkt_x(m, mkt):
        return float(xls_data["locations"]["x"]["market%s" % mkt])

    @m.Param(m.markets)
    def mkt_y(m, mkt):
        return float(xls_data["locations"]["y"]["market%s" % mkt])

    m.dist_to_mkt = Var(m.markets, bounds=(0, sqrt(300**2 + 300**2)))

    @m.Constraint(m.markets)
    def distance_calculation(m, mkt):
        return m.dist_to_mkt[mkt] == sqrt(
            (m.conv_x / 150 - m.mkt_x[mkt] / 150)**2 +
            (m.conv_y / 150 - m.mkt_y[mkt] / 150)**2) * 150

    m.shipments_to_mkt = Var(m.markets, m.quarters, bounds=(0, 400))
    m.production = Var(m.quarters, bounds=(0, 1500))

    @m.Constraint(m.quarters)
    def production_satisfaction(m, qtr):
        return m.production[qtr] == sum(m.shipments_to_mkt[mkt, qtr]
                                        for mkt in m.markets)

    @m.Constraint(m.quarters)
    def size_requirement(m, qtr):
        if qtr < m.conv_setup_time:
            return m.production[qtr] == 0
        else:
            return m.production[qtr] <= m.conv_size

    @m.Constraint(m.markets, m.quarters)
    def demand_satisfaction(m, mkt, qtr):
        return m.market_demand[mkt, qtr] <= m.shipments_to_mkt[mkt, qtr]

    m.variable_shipment_cost = Expression(
        expr=sum(
            m.shipments_to_mkt[mkt, qtr] * m.dist_to_mkt[mkt] *
            m.transport_cost[qtr]
            for mkt in m.markets for qtr in m.quarters))

    m.total_cost = Objective(
        expr=m.variable_shipment_cost + 5 * m.route_fixed_cost + m.conv_cost)

    return m


def build_modular_model():
    m = ConcreteModel()
    m.bigM = Suffix(direction=Suffix.LOCAL)
    m.quarters = RangeSet(0, 40, doc="10 years")
    m.discount_rate = Param(initialize=0.08, doc="8%")
    m.modular_setup_time = Param(initialize=1)
    m.modular_move_time = Param(initialize=1)
    m.markets = RangeSet(5)
    m.modular_sites = RangeSet(3)
    m.site_pairs = Set(
        initialize=m.modular_sites * m.modular_sites,
        filter=lambda _, x, y: not x == y)
    m.unique_site_pairs = Set(
        initialize=m.site_pairs, filter=lambda _, x, y: x < y)

    xls_data = pd.read_excel(
        os.path.join(os.path.dirname(__file__), "quarter_multiple_market_size.xlsx"),
        sheet_name=["demand", "locations"], index_col=0)

    @m.Param(m.markets, m.quarters)
    def market_demand(m, mkt, qtr):
        return float(xls_data["demand"]["market%s" % mkt][qtr])

    @m.Param(m.quarters, doc="Cost of transporting unit of product one mile.")
    def transport_cost(m, qtr):
        return 0.2 / 12 * (1 + m.discount_rate / 4) ** (-qtr / 4)

    m.route_fixed_cost = Param(
        initialize=100,
        doc="Cost of establishing a route from a modular site to a market")

    @m.Param(m.quarters, doc="Cost of transporting a module one mile")
    def modular_transport_cost(m, qtr):
        return 1 * (1 + m.discount_rate / 4) ** (-qtr / 4)
    m.modular_base_cost = Param(initialize=1000, doc="Cost for size 60")

    @m.Param(m.quarters)
    def modular_unit_cost(m, qtr):
        return m.modular_base_cost * (1 + m.discount_rate / 4) ** (-qtr / 4)

    @m.Param(m.markets)
    def mkt_x(m, mkt):
        return float(xls_data["locations"]["x"]["market%s" % mkt])

    @m.Param(m.markets)
    def mkt_y(m, mkt):
        return float(xls_data["locations"]["y"]["market%s" % mkt])

    m.site_x = Var(
        m.modular_sites, bounds=(0, 300), initialize={
            1: 50, 2: 225, 3: 250})
    m.site_y = Var(
        m.modular_sites, bounds=(0, 300), initialize={
            1: 300, 2: 275, 3: 50})

    m.num_modules = Var(
        m.modular_sites, m.quarters, domain=Integers, bounds=(0, 12))
    m.modules_transferred = Var(
        m.site_pairs, m.quarters,
        domain=Integers, bounds=(0, 12),
        doc="Number of modules moved from one site to another in a quarter.")
    # m.modules_transferred[...].fix(0)
    m.modules_added = Var(
        m.modular_sites, m.quarters, domain=Integers, bounds=(0, 12))

    m.dist_to_mkt = Var(
        m.modular_sites, m.markets, bounds=(0, sqrt(300**2 + 300**2)))
    m.sqr_scaled_dist_to_mkt = Var(
        m.modular_sites, m.markets, bounds=(0.001, 8), initialize=0.001)
    m.dist_to_site = Var(m.site_pairs, bounds=(0, sqrt(300**2 + 300**2)))
    m.sqr_scaled_dist_to_site = Var(
        m.site_pairs, bounds=(0.001, 8), initialize=0.001)
    m.shipments_to_mkt = Var(
        m.modular_sites, m.markets, m.quarters, bounds=(0, 300))
    m.production = Var(m.modular_sites, m.quarters, bounds=(0, 1500))

    @m.Disjunct(m.modular_sites)
    def site_active(disj, site):
        @disj.Constraint(m.quarters)
        def production_satisfaction(site_disj, qtr):
            return m.production[site, qtr] == sum(
                m.shipments_to_mkt[site, mkt, qtr] for mkt in m.markets)

        @disj.Constraint(m.quarters)
        def production_limit(site_disj, qtr):
            return m.production[site, qtr] <= m.num_modules[site, qtr] * 60

        @disj.Constraint(m.quarters)
        def module_balance(site_disj, qtr):
            existing_modules = m.num_modules[site, qtr - 1] if qtr >= 1 else 0
            new_modules = (m.modules_added[site, qtr - m.modular_setup_time]
                           if qtr > m.modular_setup_time else 0)
            xfrd_in_modules = sum(
                m.modules_transferred[from_site,
                                      site, qtr - m.modular_move_time]
                for from_site in m.modular_sites
                if (not from_site == site) and qtr > m.modular_move_time)
            xfrd_out_modules = sum(
                m.modules_transferred[site, to_site, qtr]
                for to_site in m.modular_sites if not to_site == site)
            return m.num_modules[site, qtr] == (
                new_modules + xfrd_in_modules - xfrd_out_modules +
                existing_modules)

    @m.Disjunct(m.modular_sites)
    def site_inactive(disj, site):
        disj.no_modules = Constraint(
            expr=sum(m.num_modules[site, qtr] for qtr in m.quarters) == 0)

        disj.no_module_transfer = Constraint(
            expr=sum(m.modules_transferred[site1, site2, qtr]
                     for site1, site2 in m.site_pairs
                     for qtr in m.quarters
                     if site1 == site or site2 == site) == 0)

        disj.no_shipments = Constraint(
            expr=sum(m.shipments_to_mkt[site, mkt, qtr]
                     for mkt in m.markets for qtr in m.quarters) == 0)

    @m.Disjunction(m.modular_sites)
    def site_active_or_not(m, site):
        return [m.site_active[site], m.site_inactive[site]]

    @m.Constraint(m.modular_sites, doc="Symmetry breaking for site activation")
    def site_active_ordering(m, site):
        if site + 1 <= max(m.modular_sites):
            return (m.site_active[site].indicator_var >=
                    m.site_active[site + 1].indicator_var)
        else:
            return Constraint.NoConstraint

    @m.Disjunct(m.unique_site_pairs)
    def pair_active(disj, site1, site2):
        disj.site1_active = Constraint(
            expr=m.site_active[site1].indicator_var == 1)
        disj.site2_active = Constraint(
            expr=m.site_active[site2].indicator_var == 1)
        disj.site_distance_calc = Constraint(
            expr=m.dist_to_site[site1, site2] == sqrt(
                m.sqr_scaled_dist_to_site[site1, site2]) * 150)
        disj.site_distance_symmetry = Constraint(
            expr=m.dist_to_site[site1, site2] == m.dist_to_site[site2, site1])
        disj.site_sqr_distance_calc = Constraint(
            expr=m.sqr_scaled_dist_to_site[site1, site2] == (
                (m.site_x[site1] / 150 - m.site_x[site2] / 150)**2 +
                (m.site_y[site1] / 150 - m.site_y[site2] / 150)**2))
        disj.site_sqr_distance_symmetry = Constraint(
            expr=m.sqr_scaled_dist_to_site[site1, site2] ==
            m.sqr_scaled_dist_to_site[site2, site1])

    @m.Disjunct(m.unique_site_pairs)
    def pair_inactive(disj, site1, site2):
        disj.site1_inactive = Constraint(
            expr=m.site_active[site1].indicator_var == 0)
        disj.site2_inactive = Constraint(
            expr=m.site_active[site2].indicator_var == 0)

        disj.no_module_transfer = Constraint(
            expr=sum(m.modules_transferred[site1, site2, qtr]
                     for qtr in m.quarters) == 0)

    @m.Disjunction(m.unique_site_pairs)
    def site_pair_active_or_not(m, site1, site2):
        return [m.pair_active[site1, site2], m.pair_inactive[site1, site2]]

    @m.Constraint(m.markets, m.quarters)
    def demand_satisfaction(m, mkt, qtr):
        return m.market_demand[mkt, qtr] <= sum(
            m.shipments_to_mkt[site, mkt, qtr] for site in m.modular_sites)

    @m.Disjunct(m.modular_sites, m.markets)
    def product_route_active(disj, site, mkt):
        disj.site_active = Constraint(
            expr=m.site_active[site].indicator_var == 1)

        @disj.Constraint()
        def market_distance_calculation(site_disj):
            return m.dist_to_mkt[site, mkt] == sqrt(
                m.sqr_scaled_dist_to_mkt[site, mkt]) * 150

        @disj.Constraint()
        def market_sqr_distance_calc(site_disj):
            return m.sqr_scaled_dist_to_mkt[site, mkt] == (
                (m.site_x[site] / 150 - m.mkt_x[mkt] / 150)**2 +
                (m.site_y[site] / 150 - m.mkt_y[mkt] / 150)**2)

    @m.Disjunct(m.modular_sites, m.markets)
    def product_route_inactive(disj, site, mkt):
        disj.no_shipments = Constraint(
            expr=sum(m.shipments_to_mkt[site, mkt, qtr]
                     for qtr in m.quarters) == 0)

    @m.Disjunction(m.modular_sites, m.markets)
    def product_route_active_or_not(m, site, mkt):
        return [m.product_route_active[site, mkt],
                m.product_route_inactive[site, mkt]]

    m.variable_shipment_cost = Expression(
        expr=sum(
            m.shipments_to_mkt[site, mkt, qtr] * m.dist_to_mkt[site, mkt] *
            m.transport_cost[qtr]
            for site in m.modular_sites for mkt in m.markets
            for qtr in m.quarters))

    m.fixed_shipment_cost = Expression(
        expr=sum(m.product_route_active[site, mkt].indicator_var *
                 m.route_fixed_cost
                 for site in m.modular_sites for mkt in m.markets))

    m.module_purchase_cost = Expression(
        expr=sum(m.modules_added[site, qtr] * m.modular_unit_cost[qtr]
                 for site in m.modular_sites for qtr in m.quarters))

    m.module_transfer_cost = Expression(expr=sum(
        m.modules_transferred[site1, site2, qtr]
        * m.dist_to_site[site1, site2] * m.modular_transport_cost[qtr]
        for site1, site2 in m.site_pairs for qtr in m.quarters))

    m.total_cost = Objective(
        expr=m.variable_shipment_cost
        + m.fixed_shipment_cost
        + m.module_purchase_cost
        + m.module_transfer_cost)

    return m


if __name__ == "__main__":
    # m = build_model()
    # SolverFactory('gams').solve(m, tee=True, solver='baron')
    # print("Plant at ({conv_x:3.0f}, {conv_y:3.0f}) size {conv_size:3.0f} "
    #       "cost {conv_cost:4.0f}.".format(
    #           conv_x=m.conv_x.value,
    #           conv_y=m.conv_y.value,
    #           conv_size=m.conv_size.value,
    #           conv_cost=m.conv_cost.value))
    # print("Total cost: %s" % value(m.total_cost.expr))
    # print("  Variable ship cost: %s" % value(m.variable_shipment_cost))
    # print("  Fix ship cost: %s" % value(500))
    # print("  Conv buy cost: %s" % value(m.conv_cost))
    # exit()

    m = build_modular_model()
    # res = SolverFactory('gdpopt').solve(
    #     m, tee=True, mip_solver="gams", nlp_solver="ipopt",
    #     mip_solver_args={'io_options': {'add_options': [
    #         'option reslim=30;']}})
    # print(res.solver.timing)
    TransformationFactory('gdp.bigm').apply_to(m, bigM=10000)
    # TransformationFactory('gdp.chull').apply_to(m)
    # res = SolverFactory('gams').solve(m, tee=True, io_options={
    #     'add_options': ['option reslim = 300;']})
    res = SolverFactory('gams').solve(
        m, tee=True,
        io_options={
            'solver': 'baron',
            'add_options': ['option reslim = 600;']})
    # from pyomo.util.infeasible import log_infeasible_constraints
    # log_infeasible_constraints(m)

    def record_generator():
        for qtr in m.quarters:
            yield (
                (qtr,)
                + tuple(m.num_modules[site, qtr].value
                        for site in m.modular_sites)
                + tuple(m.production[site, qtr].value
                        for site in m.modular_sites)
                + tuple(m.shipments_to_mkt[site, mkt, qtr].value
                        for site in m.modular_sites
                        for mkt in m.markets)
                + tuple(m.modules_added[site, qtr].value
                        for site in m.modular_sites)
                + tuple(m.modules_transferred[site1, site2, qtr].value
                        for site1, site2 in m.site_pairs)
            )

    df = pd.DataFrame.from_records(
        record_generator(),
        columns=("Qtr",)
        + tuple("Num Site%s" % site for site in m.modular_sites)
        + tuple("Prod Site%s" % site for site in m.modular_sites)
        + tuple("ShipSite%stoMkt%s" % (site, mkt)
                for site in m.modular_sites for mkt in m.markets)
        + tuple("Add Site%s" % site for site in m.modular_sites)
        + tuple("Xfer Site%s to %s" % (site1, site2)
                for site1, site2 in m.site_pairs))
    df.to_excel("quarter_multiple_modular_config.xlsx")
    print("Total cost: %s" % value(m.total_cost.expr))
    print("  Variable ship cost: %s" % value(m.variable_shipment_cost))
    print("  Fix ship cost: %s" % value(m.fixed_shipment_cost))
    print("  Module buy cost: %s" % value(m.module_purchase_cost))
    print("  Module xfer cost: %s" % value(m.module_transfer_cost))
    for site in m.modular_sites:
        print("Site {:1.0f} at ({:3.0f}, {:3.0f})".format(
            site, m.site_x[site].value, m.site_y[site].value))
        print("  Supplies markets {}".format(tuple(
            mkt for mkt in m.markets
            if m.product_route_active[site, mkt].indicator_var.value == 1)))

    if res.solver.termination_condition is not TerminationCondition.optimal:
        exit()

    plt.plot([x.value for x in m.site_x.values()],
             [y.value for y in m.site_y.values()], 'k.', markersize=12)
    plt.plot([x for x in m.mkt_x.values()],
             [y for y in m.mkt_y.values()], 'bo', markersize=12)
    for mkt in m.markets:
        plt.annotate('mkt%s' % mkt, (m.mkt_x[mkt], m.mkt_y[mkt]),
                     (m.mkt_x[mkt] + 2, m.mkt_y[mkt] + 0))
    for site in m.modular_sites:
        plt.annotate(
            'site%s' % site, (m.site_x[site].value, m.site_y[site].value),
            (m.site_x[site].value + 2, m.site_y[site].value + 0))
    for site, mkt in m.modular_sites * m.markets:
        if m.product_route_active[site, mkt].indicator_var.value == 1:
            plt.arrow(m.site_x[site].value, m.site_y[site].value,
                      m.mkt_x[mkt] - m.site_x[site].value,
                      m.mkt_y[mkt] - m.site_y[site].value,
                      width=0.8, length_includes_head=True, color='r')
    for site1, site2 in m.site_pairs:
        if sum(m.modules_transferred[site1, site2, qtr].value
               for qtr in m.quarters) > 1E-6:
            plt.arrow(m.site_x[site1].value, m.site_y[site1].value,
                      m.site_x[site2].value - m.site_x[site1].value,
                      m.site_y[site2].value - m.site_y[site1].value,
                      width=0.9, length_includes_head=True,
                      linestyle='dotted', color='k')
    plt.show()



"""Initialization routine for distillation column"""
from __future__ import division

import pandas
from math import fabs, floor
from pyomo.environ import value, exp


def initialize(m):
    m.reflux_frac.set_value(value(
        m.reflux_ratio / (1 + m.reflux_ratio)))
    m.boilup_frac.set_value(value(
        m.reboil_ratio / (1 + m.reboil_ratio)))

    _excel_sheets = pandas.read_excel('init.xlsx', sheet_name=None)

    def set_value_if_not_fixed(var, val):
        """Set variable to the value if it is not fixed."""
        if not var.fixed:
            var.set_value(val)

    active_trays = [
        t for t in m.trays
        if t not in m.conditional_trays or
        fabs(value(m.tray[t].indicator_var - 1)) <= 1E-3]
    num_active_trays = len(active_trays)

    feed_tray = m.feed_tray

    tray_indexed_data = _excel_sheets['trays']
    tray_indexed_data.sort_values(by=['tray'], inplace=True)
    tray_indexed_data.set_index('tray', inplace=True)

    comp_and_tray_indexed_data = _excel_sheets['comps_and_trays']
    comp_and_tray_indexed_data.sort_values(by=['comp', 'tray'],
                                           inplace=True)
    comp_and_tray_indexed_data.set_index(['comp', 'tray'], inplace=True)
    comp_slices = {c: comp_and_tray_indexed_data.loc[c, :]
                   for c in m.comps}

    num_data_trays = tray_indexed_data.index.size
    if num_active_trays < num_data_trays:
        # Number of model trays is less than number of trays in data. Need to
        # do averaging
        new_indices = [1] + [
            1 + (num_data_trays - 1) / (num_active_trays - 1) * i
            for i in range(1, num_active_trays)]
        for tray in range(2, num_active_trays):
            indx = new_indices[tray - 1]
            lower = floor(indx)
            frac_above = indx - lower
            # Take linear combination of values
            tray_indexed_data.loc[tray] = (
                tray_indexed_data.loc[lower] * (1 - frac_above) +
                tray_indexed_data.loc[lower + 1] * frac_above)
            for c in m.comps:
                comp_slices[c].loc[tray] = (
                    comp_slices[c].loc[lower] * (1 - frac_above) +
                    comp_slices[c].loc[lower + 1] * frac_above)
        tray_indexed_data.loc[num_active_trays] = \
            tray_indexed_data.loc[num_data_trays]
        tray_indexed_data = tray_indexed_data.head(num_active_trays)
        for c in m.comps:
            comp_slices[c].loc[num_active_trays] = \
                comp_slices[c].loc[num_data_trays]
            comp_slices[c] = comp_slices[c].head(num_active_trays)
    else:
        # Stretch the data out and do interpolation
        tray_indexed_data.index = pandas.Index(
            [1] + [int(round(num_active_trays / num_data_trays * i))
                   for i in range(2, num_data_trays + 1)], name='tray')
        tray_indexed_data = tray_indexed_data.reindex(
            [i for i in range(1, num_active_trays + 1)]).interpolate()
        for c in m.comps:
            comp_slices[c].index = pandas.Index(
                [1] + [int(round(num_active_trays / num_data_trays * i))
                       for i in range(2, num_data_trays + 1)], name='tray')
            # special handling necessary for V near top of column and L
            # near column bottom. Do not want to interpolate with one end
            # being potentially 0. (ie. V from total condenser). Instead,
            # use back fill and forward fill.
            comp_slices[c] = comp_slices[c].reindex(
                [i for i in range(1, num_active_trays + 1)])
            tray_below_condenser = sorted(active_trays, reverse=True)[1]
            if pandas.isna(comp_slices[c]['V'][tray_below_condenser]):
                # V of the tray below the condenser is N/A. Find a valid
                # value lower down to use.
                val = next(
                    comp_slices[c]['V'][t]
                    for t in reversed(list(m.trays))
                    if pandas.notna(comp_slices[c]['V'][t])
                    and not t == m.condens_tray)
                comp_slices[c]['V'][tray_below_condenser] = val
            if pandas.isna(comp_slices[c]['L'][m.reboil_tray + 1]):
                # L of the tray above the reboiler is N/A. Find a valid
                # value higher up to use.
                val = next(
                    comp_slices[c]['L'][t]
                    for t in m.trays
                    if pandas.notna(comp_slices[c]['L'][t])
                    and not t == m.reboil_tray)
                comp_slices[c]['L'][m.reboil_tray + 1] = val
            comp_slices[c] = comp_slices[c].interpolate()

    tray_indexed_data.index = pandas.Index(sorted(active_trays),
                                           name='tray')
    tray_indexed_data = tray_indexed_data.reindex(sorted(m.trays),
                                                  method='bfill')

    for t in m.trays:
        set_value_if_not_fixed(m.T[t], tray_indexed_data['T [K]'][t])

    for c in m.comps:
        comp_slices[c].index = pandas.Index(sorted(active_trays),
                                            name='tray')
        comp_slices[c] = comp_slices[c].reindex(sorted(m.trays))
        comp_slices[c][['L', 'x']] = comp_slices[c][['L', 'x']].bfill()
        comp_slices[c][['V', 'y']] = comp_slices[c][['V', 'y']].ffill()

    comp_and_tray_indexed_data = pandas.concat(comp_slices)

    for c, t in m.comps * m.trays:
        set_value_if_not_fixed(m.L[c, t],
                               comp_and_tray_indexed_data['L'][c, t])
        set_value_if_not_fixed(m.V[c, t],
                               comp_and_tray_indexed_data['V'][c, t])
        set_value_if_not_fixed(m.x[c, t],
                               comp_and_tray_indexed_data['x'][c, t])
        set_value_if_not_fixed(m.y[c, t],
                               comp_and_tray_indexed_data['y'][c, t])

    for c in m.comps:
        m.H_L_spec_feed[c].set_value(value(m.feed_liq_enthalpy_expr[c]))
        m.H_V_spec_feed[c].set_value(value(m.feed_vap_enthalpy_expr[c]))

    for t in m.trays:
        for c in m.comps:
            k = m.pvap_const[c]
            x = m.Pvap_X[c, t]

            x.set_value(value(1 - m.T[t] / k['Tc']))

            m.Pvap[c, t].set_value(value(exp((
                k['A'] * x +
                k['B'] * x ** 1.5 +
                k['C'] * x ** 3 +
                k['D'] * x ** 6) / (1 - x)) * k['Pc']))

            m.Kc[c, t].set_value(value(
                m.gamma[c, t] * m.Pvap[c, t] / m.P))

            m.H_L[c, t].set_value(value(m.liq_enthalpy_expr[t, c]))
            m.H_V[c, t].set_value(value(m.vap_enthalpy_expr[t, c]))

    m.D['benzene'].set_value(42.3152714)
    m.D['toluene'].set_value(5.4446286)
    m.B['benzene'].set_value(7.67928)
    m.B['toluene'].set_value(44.56072)
    m.L['benzene', m.reboil_tray].set_value(7.67928)
    m.L['toluene', m.reboil_tray].set_value(44.56072)
    m.V['benzene', m.reboil_tray].set_value(value(
        m.L['benzene', m.reboil_tray + 1] -
        m.L['benzene', m.reboil_tray]))
    m.V['toluene', m.reboil_tray].set_value(value(
        m.L['toluene', m.reboil_tray + 1] -
        m.L['toluene', m.reboil_tray]))
    m.L['benzene', m.condens_tray].set_value(value(
        m.V['benzene', m.condens_tray - 1] -
        m.D['benzene']))
    m.L['toluene', m.condens_tray].set_value(value(
        m.V['toluene', m.condens_tray - 1] -
        m.D['toluene']))

    for t in m.trays:
        m.liq[t].set_value(value(sum(m.L[c, t] for c in m.comps)))
        m.vap[t].set_value(value(sum(m.V[c, t] for c in m.comps)))
    m.bot.set_value(52.24)
    m.dis.set_value(47.7599)
    for c in m.comps:
        m.x[c, m.reboil_tray].set_value(value(
            m.L[c, m.reboil_tray] / m.liq[m.reboil_tray]))
        m.y[c, m.reboil_tray].set_value(value(
            m.V[c, m.reboil_tray] / m.vap[m.reboil_tray]))
        m.x[c, m.condens_tray].set_value(value(
            m.L[c, m.condens_tray] / m.liq[m.condens_tray]))
        m.y[c, m.condens_tray].set_value(value(
            m.x[c, m.condens_tray] * m.Kc[c, m.condens_tray]))
    m.Qb.set_value(2.307873115)
    m.Qc.set_value(3.62641882)



from __future__ import division

import os
from math import fabs

import pandas as pd
from pyomo.environ import (
    ConcreteModel, Constraint, Integers, minimize, NonNegativeReals, Objective, Param, RangeSet, SolverFactory, sqrt,
    Suffix, summation, TransformationFactory, value, Var, )
from pyomo.gdp import Disjunct


def build_model():
    m = ConcreteModel()
    m.bigM = Suffix(direction=Suffix.LOCAL)
    m.time = RangeSet(0, 120, doc="months in 10 years")
    m.suppliers = RangeSet(10)
    m.markets = RangeSet(10)
    m.potential_sites = RangeSet(12)
    m.discount_rate = Param(initialize=0.08, doc="8%")
    m.conv_setup_time = Param(initialize=12)
    m.modular_setup_time = Param(initialize=3)
    m.modular_teardown_time = Param(initialize=3)
    m.teardown_value = Param(initialize=0.30, doc="30%")
    m.conventional_salvage_value = Param(initialize=0.05, doc="5%")

    @m.Param(m.time)
    def discount_factor(m, t):
        return (1 + m.discount_rate / 12) ** (-t / 12)

    xls_data = pd.read_excel(
        os.path.join(os.path.dirname(__file__), "problem_data.xlsx"),
        sheet_name=["sources", "markets", "sites", "growth", "decay"],
        index_col=0)

    @m.Param(m.markets, m.time, doc="Market demand [thousand ton/month]")
    def market_demand(m, mkt, t):
        if m.conv_setup_time <= t <= max(m.time) - 3:
            return float(xls_data["markets"]["demand"][mkt]) / 1000 / 12
        else:
            return 0

    @m.Param(m.suppliers, m.time, doc="Raw material supply [thousand ton/month]")
    def available_supply(m, sup, t):
        # If t is before supply available or after supply decayed, then no
        # supply
        if t < float(xls_data["sources"]["growth"][sup]):
            return 0
        elif t > float(xls_data["sources"]["decay"][sup]):
            return 0
        else:
            return float(xls_data["sources"]["avail"][sup]) / 1000 / 12

    @m.Param(m.suppliers)
    def supplier_x(m, sup):
        return float(xls_data["sources"]["x"][sup])

    @m.Param(m.suppliers)
    def supplier_y(m, sup):
        return float(xls_data["sources"]["y"][sup])

    @m.Param(m.markets)
    def market_x(m, mkt):
        return float(xls_data["markets"]["x"][mkt])

    @m.Param(m.markets)
    def market_y(m, mkt):
        return float(xls_data["markets"]["y"][mkt])

    @m.Param(m.potential_sites)
    def site_x(m, site):
        return float(xls_data["sites"]["x"][site])

    @m.Param(m.potential_sites)
    def site_y(m, site):
        return float(xls_data["sites"]["y"][site])

    @m.Param(m.suppliers, m.potential_sites, doc="Miles")
    def dist_supplier_to_site(m, sup, site):
        return sqrt((m.supplier_x[sup] - m.site_x[site]) ** 2 +
                    (m.supplier_y[sup] - m.site_y[site]) ** 2)

    @m.Param(m.potential_sites, m.markets, doc="Miles")
    def dist_site_to_market(m, site, mkt):
        return sqrt((m.site_x[site] - m.market_x[mkt]) ** 2 +
                    (m.site_y[site] - m.market_y[mkt]) ** 2)

    m.conversion = Param(initialize=0.26, doc="overall conversion to product")

    m.conv_site_size = Var(
        m.potential_sites,
        bounds=(120 / 12 / 10, 120 / 12), initialize=1,
        doc="Product capacity of site [thousand ton/mo]")

    m.conv_base_cost = Param(initialize=268.4, doc="Cost for size 120k per year [million $]")
    m.module_base_cost = Param(initialize=268.4, doc="Cost for size 120k per year [million $]")
    m.conv_exponent = Param(initialize=0.7)

    m.supply = Var(m.potential_sites, m.time, bounds=(0, 120 / 12 / 0.26 * 10), doc="thousand ton/mo")
    m.production = Var(m.potential_sites, m.time, bounds=(0, 120 / 12 * 10), doc="thousand ton/mo")
    m.num_modules = Var(m.potential_sites, m.time, domain=Integers, bounds=(0, 10))
    m.modules_purchased = Var(m.potential_sites, m.time, domain=Integers, bounds=(0, 10))
    m.modules_sold = Var(m.potential_sites, m.time, domain=Integers, bounds=(0, 10))

    m.conv_build_cost = Var(
        m.potential_sites,
        doc="Cost of building conventional facility [milllion $]",
        bounds=(0, 1350 * 10), initialize=0)

    @m.Param(m.suppliers, m.time)
    def raw_material_unit_cost(m, sup, t):
        return float(xls_data["sources"]["cost"][sup]) * m.discount_factor[t]

    @m.Param(m.time)
    def module_unit_cost(m, t):
        return m.module_base_cost * m.discount_factor[t]

    @m.Param(m.time, doc="$/ton")
    def unit_production_cost(m, t):
        return 300 * m.discount_factor[t]

    @m.Param(doc="thousand $")
    def transport_fixed_cost(m):
        return 125

    @m.Param(m.time, doc="$/ton-mile")
    def unit_product_transport_cost(m, t):
        return 0.13 * m.discount_factor[t]

    @m.Param(m.time, doc="$/ton-mile")
    def unit_raw_material_transport_cost(m, t):
        return 2 * m.discount_factor[t]

    m.supply_shipments = Var(
        m.suppliers, m.potential_sites, m.time, domain=NonNegativeReals,
        bounds=(0, 120 / 12 / 0.26), doc="thousand ton/mo")
    m.product_shipments = Var(
        m.potential_sites, m.markets, m.time, domain=NonNegativeReals,
        bounds=(0, 120 / 12), doc="thousand ton/mo")

    @m.Constraint(m.suppliers, m.time)
    def supply_limits(m, sup, t):
        return sum(m.supply_shipments[sup, site, t]
                   for site in m.potential_sites) <= m.available_supply[sup, t]

    @m.Constraint(m.markets, m.time)
    def demand_satisfaction(m, mkt, t):
        return sum(m.product_shipments[site, mkt, t]
                   for site in m.potential_sites) == m.market_demand[mkt, t]

    @m.Constraint(m.potential_sites, m.time)
    def product_balance(m, site, t):
        return m.production[site, t] == sum(m.product_shipments[site, mkt, t]
                                            for mkt in m.markets)

    @m.Constraint(m.potential_sites, m.time)
    def require_raw_materials(m, site, t):
        return m.production[site, t] <= m.conversion * m.supply[site, t]

    m.modular = Disjunct(m.potential_sites, rule=_build_modular_disjunct)
    m.conventional = Disjunct(m.potential_sites, rule=_build_conventional_disjunct)
    m.site_inactive = Disjunct(m.potential_sites, rule=_build_site_inactive_disjunct)

    m.supply_route_active = Disjunct(m.suppliers, m.potential_sites, rule=_build_supply_route_active)
    m.supply_route_inactive = Disjunct(m.suppliers, m.potential_sites, rule=_build_supply_route_inactive)

    m.product_route_active = Disjunct(m.potential_sites, m.markets, rule=_build_product_route_active)
    m.product_route_inactive = Disjunct(m.potential_sites, m.markets, rule=_build_product_route_inactive)

    @m.Disjunction(m.potential_sites)
    def site_type(m, site):
        return [m.modular[site], m.conventional[site], m.site_inactive[site]]

    @m.Disjunction(m.suppliers, m.potential_sites)
    def supply_route_active_or_not(m, sup, site):
        return [m.supply_route_active[sup, site], m.supply_route_inactive[sup, site]]

    @m.Disjunction(m.potential_sites, m.markets)
    def product_route_active_or_not(m, site, mkt):
        return [m.product_route_active[site, mkt], m.product_route_inactive[site, mkt]]

    @m.Expression(m.suppliers, m.potential_sites, doc="million $")
    def raw_material_transport_cost(m, sup, site):
        return sum(
            m.supply_shipments[sup, site, t]
            * m.unit_raw_material_transport_cost[t]
            * m.dist_supplier_to_site[sup, site] / 1000
            for t in m.time)

    @m.Expression(doc="million $")
    def raw_material_fixed_transport_cost(m):
        return (
            sum(m.supply_route_active[sup, site].indicator_var
                for sup in m.suppliers for site in m.potential_sites)
            * m.transport_fixed_cost / 1000)

    @m.Expression(m.potential_sites, m.markets, doc="million $")
    def product_transport_cost(m, site, mkt):
        return sum(
            m.product_shipments[site, mkt, t]
            * m.unit_product_transport_cost[t]
            * m.dist_site_to_market[site, mkt] / 1000
            for t in m.time)

    @m.Expression(doc="million $")
    def product_fixed_transport_cost(m):
        return (
            sum(m.product_route_active[site, mkt].indicator_var
                for site in m.potential_sites for mkt in m.markets)
            * m.transport_fixed_cost / 1000)

    @m.Expression(m.potential_sites, m.time, doc="Cost of module setups in each month [million $]")
    def module_setup_cost(m, site, t):
        return m.modules_purchased[site, t] * m.module_unit_cost[t]

    @m.Expression(m.potential_sites, m.time, doc="Value of module teardowns in each month [million $]")
    def module_teardown_credit(m, site, t):
        return m.modules_sold[site, t] * m.module_unit_cost[t] * m.teardown_value

    @m.Expression(m.potential_sites, doc="Conventional site salvage value")
    def conv_salvage_value(m, site):
        return m.conv_build_cost[site] * m.discount_factor[m.time.last()] * m.conventional_salvage_value

    m.total_cost = Objective(
        expr=0
        + summation(m.conv_build_cost)
        + summation(m.module_setup_cost)
        - summation(m.conv_salvage_value)
        - summation(m.module_teardown_credit)
        + summation(m.raw_material_transport_cost)
        + summation(m.raw_material_fixed_transport_cost)
        + summation(m.product_transport_cost)
        + summation(m.product_fixed_transport_cost)
        + 0,
        sense=minimize)

    return m


def _build_site_inactive_disjunct(disj, site):
    m = disj.model()

    @disj.Constraint()
    def no_modules(disj):
        return sum(m.num_modules[...]) + sum(m.modules_purchased[...]) + sum(m.modules_sold[...]) == 0

    @disj.Constraint()
    def no_production(disj):
        return sum(m.production[site, t] for t in m.time) == 0

    @disj.Constraint()
    def no_supply(disj):
        return sum(m.supply[site, t] for t in m.time) == 0


def _build_conventional_disjunct(disj, site):
    m = disj.model()

    disj.cost_calc = Constraint(
        expr=m.conv_build_cost[site] == (
            m.conv_base_cost * (m.conv_site_size[site] / 10) ** m.conv_exponent))
    # m.bigM[disj.cost_calc] = 7000

    @disj.Constraint(m.time)
    def supply_balance(disj, t):
        return m.supply[site, t] == sum(
            m.supply_shipments[sup, site, t] for sup in m.suppliers)

    @disj.Constraint(m.time)
    def conv_production_limit(conv_disj, t):
        if t < m.conv_setup_time:
            return m.production[site, t] == 0
        else:
            return m.production[site, t] <= m.conv_site_size[site]

    @disj.Constraint()
    def no_modules(disj):
        return sum(m.num_modules[...]) + sum(m.modules_purchased[...]) + sum(m.modules_sold[...]) == 0


def _build_modular_disjunct(disj, site):
    m = disj.model()

    @disj.Constraint(m.time)
    def supply_balance(disj, t):
        return m.supply[site, t] == sum(
            m.supply_shipments[sup, site, t] for sup in m.suppliers)

    @disj.Constraint(m.time)
    def module_balance(disj, t):
        existing_modules = 0 if t == m.time.first() else m.num_modules[site, t - 1]
        new_modules = 0 if t < m.modular_setup_time else m.modules_purchased[site, t - m.modular_setup_time]
        sold_modules = m.modules_sold[site, t]
        return m.num_modules[site, t] == existing_modules + new_modules - sold_modules

    for t in range(value(m.modular_setup_time)):
        m.num_modules[site, t].fix(0)

    @disj.Constraint(m.time)
    def modular_production_limit(mod_disj, t):
        return m.production[site, t] <= 10 * m.num_modules[site, t]


def _build_supply_route_active(disj, sup, site):
    m = disj.model()


def _build_supply_route_inactive(disj, sup, site):
    m = disj.model()

    @disj.Constraint()
    def no_supply(disj):
        return sum(m.supply_shipments[sup, site, t] for t in m.time) == 0


def _build_product_route_active(disj, site, mkt):
    m = disj.model()


def _build_product_route_inactive(disj, site, mkt):
    m = disj.model()

    @disj.Constraint()
    def no_product(disj):
        return sum(m.product_shipments[site, mkt, t] for t in m.time) == 0


def print_nonzeros(var):
    for i in var:
        if var[i].value != 0:
            print("%7s : %10f : %10f : %10f" % (i, var[i].lb, var[i].value, var[i].ub))


if __name__ == "__main__":
    m = build_model()
    m.modular[:].deactivate()
    # m.conventional[:].deactivate()
    # for site, t in m.potential_sites * m.time:
    #     if t <= max(m.time) - 4:
    #         m.modular[site].removing_modules[t].deactivate()
    TransformationFactory('gdp.bigm').apply_to(m, bigM=7000)
    # res = SolverFactory('gurobi').solve(m, tee=True)
    res = SolverFactory('gams').solve(
        m, tee=True,
        solver='scip',
        # solver='gurobi',
        # add_options=['option reslim = 1200;', 'option optcr=0.0001;'],
        add_options=[
            'option reslim = 1200;',
            'OPTION threads=4;',
            'option optcr=0.01',
            ],
        )
    # res = SolverFactory('gdpopt').solve(
    #     m, tee=True,
    #     iterlim=2,
    #     mip_solver='gams',
    #     mip_solver_args=dict(add_options=['option reslim = 30;']))
    results = pd.DataFrame([
        ['Total Cost', value(m.total_cost)],
        ['Conv Build Cost', value(summation(m.conv_build_cost))],
        ['Conv Salvage Value', value(summation(m.conv_salvage_value))],
        ['Module Build Cost', value(summation(m.module_setup_cost))],
        ['Module Salvage Value', value(summation(m.module_teardown_credit))],
        ['Raw Material Transport', value(summation(m.raw_material_transport_cost) + summation(m.raw_material_fixed_transport_cost))],
        ['Product Transport', value(summation(m.product_transport_cost) + summation(m.product_fixed_transport_cost))]
    ], columns=['Quantity', 'Value [million $]']).set_index('Quantity').round(0)
    print(results)

    df = pd.DataFrame([
        [
            site, t,
            value(m.num_modules[site, t]),
            value(m.modules_purchased[site, t]),
            value(m.modules_sold[site, t]),
            value(m.module_setup_cost[site, t]),
            value(m.module_teardown_credit[site, t]),
            value(m.production[site, t])] for site, t in m.potential_sites * m.time
        ],
        columns=("Site", "Month", "Num Modules", "Buy Modules",
                 "Sell Modules",
                 "Setup Cost", "Teardown Credit", "Production")
    )
    df.to_excel("facility_config.xlsx")

    # if res.solver.termination_condition is not TerminationCondition.optimal:
    #     exit()
    import matplotlib.pyplot as plt

    plt.plot([x for x in m.site_x.values()],
             [y for y in m.site_y.values()], 'k.', markersize=12)
    plt.plot([x for x in m.market_x.values()],
             [y for y in m.market_y.values()], 'b.', markersize=12)
    plt.plot([x for x in m.supplier_x.values()],
             [y for y in m.supplier_y.values()], 'r.', markersize=12)
    for mkt in m.markets:
        plt.annotate('m%s' % mkt, (m.market_x[mkt], m.market_y[mkt]),
                     (m.market_x[mkt] + 2, m.market_y[mkt] + 2),
                     fontsize='x-small')
    for site in m.potential_sites:
        if m.site_inactive[site].indicator_var.value == 0:
            plt.annotate(
                'p%s' % site, (m.site_x[site], m.site_y[site]),
                (m.site_x[site] + 2, m.site_y[site] + 2),
                fontsize='x-small')
        else:
            plt.annotate(
                'x%s' % site, (m.site_x[site], m.site_y[site]),
                (m.site_x[site] + 2, m.site_y[site] + 2),
                fontsize='x-small')
    for sup in m.suppliers:
        plt.annotate(
            's%s' % sup, (m.supplier_x[sup], m.supplier_y[sup]),
            (m.supplier_x[sup] + 2, m.supplier_y[sup] + 2),
            fontsize='x-small')
    for sup, site in m.suppliers * m.potential_sites:
        if fabs(m.supply_route_active[sup, site].indicator_var.value - 1) <= 1E-3:
            plt.arrow(m.supplier_x[sup], m.supplier_y[sup],
                      m.site_x[site] - m.supplier_x[sup],
                      m.site_y[site] - m.supplier_y[sup],
                      width=0.8, length_includes_head=True, color='b')
    for site, mkt in m.potential_sites * m.markets:
        if fabs(m.product_route_active[site, mkt].indicator_var.value - 1) <= 1E-3:
            plt.arrow(m.site_x[site], m.site_y[site],
                      m.market_x[mkt] - m.site_x[site],
                      m.market_y[mkt] - m.site_y[site],
                      width=0.8, length_includes_head=True, color='r')
    plt.show()



"""Distillation column model for 2018 PSE conference"""

from __future__ import division

from pyomo.environ import (
    Block, ConcreteModel, Constraint, log, minimize, NonNegativeReals, Objective, RangeSet, Set, Var, )
from pyomo.gdp import Disjunct, Disjunction


def build_column(min_trays, max_trays, xD, xB):
    """Builds the column model."""
    m = ConcreteModel('benzene-toluene column')
    m.comps = Set(initialize=['benzene', 'toluene'])
    min_T, max_T = 300, 400
    max_flow = 500
    m.T_feed = Var(
        doc='Feed temperature [K]', domain=NonNegativeReals,
        bounds=(min_T, max_T), initialize=368)
    m.feed_vap_frac = Var(
        doc='Vapor fraction of feed',
        initialize=0, bounds=(0, 1))
    m.feed = Var(
        m.comps, doc='Total component feed flow [mol/s]', initialize=50)

    m.condens_tray = max_trays
    m.feed_tray = int(round(max_trays / 2))
    m.reboil_tray = 1
    m.distillate_purity = xD
    m.bottoms_purity = xB

    m.trays = RangeSet(max_trays, doc='Set of potential trays')
    m.conditional_trays = Set(
        initialize=m.trays - [m.condens_tray, m.feed_tray, m.reboil_tray],
        doc="Trays that may be turned on and off.")
    m.tray = Disjunct(m.conditional_trays, doc='Disjunct for tray existence')
    m.no_tray = Disjunct(m.conditional_trays, doc='Disjunct for tray absence')

    @m.Disjunction(m.conditional_trays, doc='Tray exists or does not')
    def tray_no_tray(b, t):
        return [b.tray[t], b.no_tray[t]]
    m.minimum_num_trays = Constraint(
        expr=sum(m.tray[t].indicator_var
                 for t in m.conditional_trays) + 1  # for feed tray
        >= min_trays)

    m.x = Var(m.comps, m.trays, doc='Liquid mole fraction',
              bounds=(0, 1), domain=NonNegativeReals, initialize=0.5)
    m.y = Var(m.comps, m.trays, doc='Vapor mole fraction',
              bounds=(0, 1), domain=NonNegativeReals, initialize=0.5)
    m.L = Var(m.comps, m.trays,
              doc='component liquid flows from tray in kmol',
              domain=NonNegativeReals, bounds=(0, max_flow),
              initialize=50)
    m.V = Var(m.comps, m.trays,
              doc='component vapor flows from tray in kmol',
              domain=NonNegativeReals, bounds=(0, max_flow),
              initialize=50)
    m.liq = Var(m.trays, domain=NonNegativeReals,
                doc='liquid flows from tray in kmol', initialize=100,
                bounds=(0, max_flow))
    m.vap = Var(m.trays, domain=NonNegativeReals,
                doc='vapor flows from tray in kmol', initialize=100,
                bounds=(0, max_flow))
    m.B = Var(m.comps, domain=NonNegativeReals,
              doc='bottoms component flows in kmol',
              bounds=(0, max_flow), initialize=50)
    m.D = Var(m.comps, domain=NonNegativeReals,
              doc='distillate component flows in kmol',
              bounds=(0, max_flow), initialize=50)
    m.bot = Var(domain=NonNegativeReals, initialize=50, bounds=(0, 100),
                doc='bottoms flow in kmol')
    m.dis = Var(domain=NonNegativeReals, initialize=50,
                doc='distillate flow in kmol', bounds=(0, 100))
    m.reflux_ratio = Var(domain=NonNegativeReals, bounds=(0.5, 4),
                         doc='reflux ratio', initialize=0.8329)
    m.reboil_ratio = Var(domain=NonNegativeReals, bounds=(0.5, 4),
                         doc='reboil ratio', initialize=0.9527)
    m.reflux_frac = Var(domain=NonNegativeReals, bounds=(0, 1 - 1E-6),
                        doc='reflux fractions')
    m.boilup_frac = Var(domain=NonNegativeReals, bounds=(0, 1 - 1E-6),
                        doc='boilup fraction')

    m.partial_cond = Disjunct()
    m.total_cond = Disjunct()
    m.condenser_choice = Disjunction(expr=[m.partial_cond, m.total_cond])

    for t in m.conditional_trays:
        _build_conditional_tray_mass_balance(m, t, m.tray[t], m.no_tray[t])
    _build_feed_tray_mass_balance(m)
    _build_condenser_mass_balance(m)
    _build_reboiler_mass_balance(m)

    @m.Constraint(m.comps,
                  doc="Bottoms flow is equal to liquid leaving reboiler.")
    def bottoms_mass_balance(m, c):
        return m.B[c] == m.L[c, m.reboil_tray]

    @m.Constraint()
    def boilup_frac_defn(m):
        return m.bot == (1 - m.boilup_frac) * m.liq[m.reboil_tray + 1]

    @m.Constraint()
    def reflux_frac_defn(m):
        return m.dis == (1 - m.reflux_frac) * (
            m.vap[m.condens_tray - 1] - m.vap[m.condens_tray])

    @m.Constraint(m.trays)
    def liquid_sum(m, t):
        return sum(m.L[c, t] for c in m.comps) == m.liq[t]

    @m.Constraint(m.trays)
    def vapor_sum(m, t):
        return sum(m.V[c, t] for c in m.comps) == m.vap[t]

    m.bottoms_sum = Constraint(
        expr=sum(m.B[c] for c in m.comps) == m.bot)
    m.distil_sum = Constraint(
        expr=sum(m.D[c] for c in m.comps) == m.dis)

    """Phase Equilibrium relations"""
    m.Kc = Var(
        m.comps, m.trays, doc='Phase equilibrium constant',
        domain=NonNegativeReals, initialize=1, bounds=(0, 1000))
    m.T = Var(m.trays, doc='Temperature [K]',
              domain=NonNegativeReals,
              bounds=(min_T, max_T))

    @m.Constraint(m.trays)
    def monotonoic_temperature(_, t):
        return m.T[t] >= m.T[t + 1] if t < max_trays else Constraint.Skip

    m.P = Var(doc='Pressure [bar]',
              bounds=(0, 5))
    m.P.fix(1.01)
    m.T_ref = 298.15
    m.gamma = Var(
        m.comps, m.trays,
        doc='liquid activity coefficent of component on tray',
        domain=NonNegativeReals, bounds=(0, 10), initialize=1)
    m.Pvap = Var(
        m.comps, m.trays,
        doc='pure component vapor pressure of component on tray in bar',
        domain=NonNegativeReals, bounds=(1E-3, 5), initialize=0.4)
    m.Pvap_X = Var(
        m.comps, m.trays,
        doc='Related to fraction of critical temperature (1 - T/Tc)',
        bounds=(0.25, 0.5), initialize=0.4)

    m.pvap_const = {
        'benzene': {'A': -6.98273, 'B': 1.33213, 'C': -2.62863,
                    'D': -3.33399, 'Tc': 562.2, 'Pc': 48.9},
        'toluene': {'A': -7.28607, 'B': 1.38091, 'C': -2.83433,
                    'D': -2.79168, 'Tc': 591.8, 'Pc': 41.0}}

    for t in m.conditional_trays:
        _build_tray_phase_equilibrium(m, t, m.tray[t])
    m.feed_tray_phase_eq = Block()
    m.reboiler_phase_eq = Block()
    m.condenser_phase_eq = Block()
    _build_tray_phase_equilibrium(m, m.feed_tray, m.feed_tray_phase_eq)
    _build_tray_phase_equilibrium(m, m.reboil_tray, m.reboiler_phase_eq)
    _build_tray_phase_equilibrium(m, m.condens_tray, m.condenser_phase_eq)

    m.H_L = Var(
        m.comps, m.trays, bounds=(0.1, 16),
        doc='Liquid molar enthalpy of component in tray (kJ/mol)')
    m.H_V = Var(
        m.comps, m.trays, bounds=(30, 16 + 40),
        doc='Vapor molar enthalpy of component in tray (kJ/mol)')
    m.H_L_spec_feed = Var(
        m.comps, doc='Component liquid molar enthalpy in feed [kJ/mol]',
        initialize=0, bounds=(0.1, 16))
    m.H_V_spec_feed = Var(
        m.comps, doc='Component vapor molar enthalpy in feed [kJ/mol]',
        initialize=0, bounds=(30, 16 + 40))
    m.Qb = Var(domain=NonNegativeReals, doc='reboiler duty (MJ/s)',
               initialize=1, bounds=(0, 8))
    m.Qc = Var(domain=NonNegativeReals, doc='condenser duty (MJ/s)',
               initialize=1, bounds=(0, 8))

    m.vap_Cp_const = {
        'benzene': {'A': -3.392E1, 'B': 4.739E-1, 'C': -3.017E-4,
                    'D': 7.130E-8, 'E': 0},
        'toluene': {'A': -2.435E1, 'B': 5.125E-1, 'C': -2.765E-4,
                    'D': 4.911E-8, 'E': 0}}
    m.liq_Cp_const = {
        'benzene': {'A': 1.29E5, 'B': -1.7E2, 'C': 6.48E-1,
                    'D': 0, 'E': 0},
        'toluene': {'A': 1.40E5, 'B': -1.52E2, 'C': 6.95E-1,
                    'D': 0, 'E': 0}}
    m.dH_vap = {'benzene': 33.770E3, 'toluene': 38.262E3}  # J/mol

    _build_column_heat_relations(m)

    @m.Constraint()
    def distillate_req(m):
        return m.D['benzene'] >= m.distillate_purity * m.dis

    @m.Constraint()
    def bottoms_req(m):
        return m.B['toluene'] >= m.bottoms_purity * m.bot

    # m.obj = Objective(expr=(m.Qc + m.Qb) * 1E-3, sense=minimize)
    m.obj = Objective(
        expr=(m.Qc + m.Qb) * 1E3 + 1E3 * (
            sum(m.tray[t].indicator_var for t in m.conditional_trays) + 1),
        sense=minimize)
    # m.obj = Objective(
    #     expr=sum(m.tray[t].indicator_var for t in m.conditional_trays) + 1)

    @m.Constraint()
    def reflux_ratio_calc(m):
        return m.reflux_frac * (m.reflux_ratio + 1) == m.reflux_ratio

    @m.Constraint()
    def reboil_ratio_calc(m):
        return m.boilup_frac * (m.reboil_ratio + 1) == m.reboil_ratio

    @m.Constraint(m.conditional_trays)
    def tray_ordering(m, t):
        """Trays close to the feed should be activated first."""
        if t + 1 < m.condens_tray and t > m.feed_tray:
            return m.tray[t].indicator_var >= m.tray[t + 1].indicator_var
        elif t > m.reboil_tray and t + 1 < m.feed_tray:
            return m.tray[t + 1].indicator_var >= m.tray[t].indicator_var
        else:
            return Constraint.NoConstraint

    return m


def _build_conditional_tray_mass_balance(m, t, tray, no_tray):
    """
    t = tray number
    tray = tray exists disjunct
    no_tray = tray absent disjunct
    """
    @tray.Constraint(m.comps)
    def mass_balance(_, c):
        return (
            # Feed in if feed tray
            (m.feed[c] if t == m.feed_tray else 0)
            # Vapor from tray t
            - m.V[c, t]
            # Loss to distillate if condenser
            - (m.D[c] if t == m.condens_tray else 0)
            # Liquid from tray above if not condenser
            + (m.L[c, t + 1] if t < m.condens_tray else 0)
            # Loss to bottoms if reboiler
            - (m.B[c] if t == m.reboil_tray else 0)
            # Liquid to tray below if not reboiler
            - (m.L[c, t] if t > m.reboil_tray else 0)
            # Vapor from tray below if not reboiler
            + (m.V[c, t - 1] if t > m.reboil_tray else 0) == 0)

    @tray.Constraint(m.comps)
    def tray_liquid_composition(_, c):
        return m.L[c, t] == m.liq[t] * m.x[c, t]

    @tray.Constraint(m.comps)
    def tray_vapor_compositions(_, c):
        return m.V[c, t] == m.vap[t] * m.y[c, t]

    @no_tray.Constraint(m.comps)
    def liq_comp_pass_through(_, c):
        return m.x[c, t] == m.x[c, t + 1]

    @no_tray.Constraint(m.comps)
    def liq_flow_pass_through(_, c):
        return m.L[c, t] == m.L[c, t + 1]

    @no_tray.Constraint(m.comps)
    def vap_comp_pass_through(_, c):
        return m.y[c, t] == m.y[c, t - 1]

    @no_tray.Constraint(m.comps)
    def vap_flow_pass_through(_, c):
        return m.V[c, t] == m.V[c, t - 1]


def _build_feed_tray_mass_balance(m):
    t = m.feed_tray

    @m.Constraint(m.comps)
    def feed_mass_balance(_, c):
        return (
            m.feed[c]        # Feed in
            - m.V[c, t]      # Vapor from tray t
            + m.L[c, t + 1]  # Liquid from tray above
            - m.L[c, t]      # Liquid to tray below
            + m.V[c, t - 1]  # Vapor from tray below
            == 0)

    @m.Constraint(m.comps)
    def feed_tray_liquid_composition(_, c):
        return m.L[c, t] == m.liq[t] * m.x[c, t]

    @m.Constraint(m.comps)
    def feed_tray_vapor_composition(_, c):
        return m.V[c, t] == m.vap[t] * m.y[c, t]


def _build_condenser_mass_balance(m):
    t = m.condens_tray

    @m.Constraint(m.comps)
    def condenser_mass_balance(_, c):
        return (
            - m.V[c, t]      # Vapor from tray t
            - m.D[c]         # Loss to distillate
            - m.L[c, t]      # Liquid to tray below
            + m.V[c, t - 1]  # Vapor from tray below
            == 0)

    @m.partial_cond.Constraint(m.comps)
    def condenser_liquid_composition(_, c):
        return m.L[c, t] == m.liq[t] * m.x[c, t]

    @m.partial_cond.Constraint(m.comps)
    def condenser_vapor_composition(_, c):
        return m.V[c, t] == m.vap[t] * m.y[c, t]

    @m.total_cond.Constraint(m.comps)
    def no_vapor_flow(_, c):
        return m.V[c, t] == 0

    @m.total_cond.Constraint()
    def no_total_vapor_flow(_):
        return m.vap[t] == 0

    @m.total_cond.Constraint(m.comps)
    def liquid_fraction_pass_through(_, c):
        return m.x[c, t] == m.y[c, t - 1]

    @m.Constraint(m.comps)
    def condenser_distillate_composition(_, c):
        return m.D[c] == m.dis * m.x[c, t]


def _build_reboiler_mass_balance(m):
    t = m.reboil_tray

    @m.Constraint(m.comps)
    def reboiler_mass_balance(_, c):
        t = m.reboil_tray
        return (
            - m.V[c, t]      # Vapor from tray t
            + m.L[c, t + 1]  # Liquid from tray above
            - m.B[c]         # Loss to bottoms
            == 0)

    @m.Constraint(m.comps)
    def reboiler_liquid_composition(_, c):
        return m.L[c, t] == m.liq[t] * m.x[c, t]

    @m.Constraint(m.comps)
    def reboiler_vapor_composition(_, c):
        return m.V[c, t] == m.vap[t] * m.y[c, t]


def _build_tray_phase_equilibrium(m, t, tray):
    @tray.Constraint(m.comps)
    def raoults_law(_, c):
        return m.y[c, t] == m.x[c, t] * m.Kc[c, t]

    @tray.Constraint(m.comps)
    def phase_equil_const(_, c):
        return m.Kc[c, t] * m.P == (
            m.gamma[c, t] * m.Pvap[c, t])

    @tray.Constraint(m.comps)
    def Pvap_relation(_, c):
        k = m.pvap_const[c]
        x = m.Pvap_X[c, t]
        return (log(m.Pvap[c, t]) - log(k['Pc'])) * (1 - x) == (
            k['A'] * x +
            k['B'] * x ** 1.5 +
            k['C'] * x ** 3 +
            k['D'] * x ** 6)

    @tray.Constraint(m.comps)
    def Pvap_X_defn(_, c):
        k = m.pvap_const[c]
        return m.Pvap_X[c, t] == 1 - m.T[t] / k['Tc']

    @tray.Constraint(m.comps)
    def gamma_calc(_, c):
        return m.gamma[c, t] == 1


def _build_column_heat_relations(m):
    @m.Expression(m.trays, m.comps)
    def liq_enthalpy_expr(_, t, c):
        k = m.liq_Cp_const[c]
        return (
            k['A'] * (m.T[t] - m.T_ref) +
            k['B'] * (m.T[t] ** 2 - m.T_ref ** 2) / 2 +
            k['C'] * (m.T[t] ** 3 - m.T_ref ** 3) / 3 +
            k['D'] * (m.T[t] ** 4 - m.T_ref ** 4) / 4 +
            k['E'] * (m.T[t] ** 5 - m.T_ref ** 5) / 5) * 1E-6

    @m.Expression(m.trays, m.comps)
    def vap_enthalpy_expr(_, t, c):
        k = m.vap_Cp_const[c]
        return (
            m.dH_vap[c] +
            k['A'] * (m.T[t] - m.T_ref) +
            k['B'] * (m.T[t] ** 2 - m.T_ref ** 2) / 2 +
            k['C'] * (m.T[t] ** 3 - m.T_ref ** 3) / 3 +
            k['D'] * (m.T[t] ** 4 - m.T_ref ** 4) / 4 +
            k['E'] * (m.T[t] ** 5 - m.T_ref ** 5) / 5) * 1E-3

    for t in m.conditional_trays:
        _build_conditional_tray_energy_balance(m, t, m.tray[t], m.no_tray[t])
    _build_feed_tray_energy_balance(m)
    _build_condenser_energy_balance(m)
    _build_reboiler_energy_balance(m)


def _build_conditional_tray_energy_balance(m, t, tray, no_tray):
    @tray.Constraint()
    def energy_balance(_):
        return sum(
            m.L[c, t + 1] * m.H_L[c, t + 1]  # heat of liquid from tray above
            - m.L[c, t] * m.H_L[c, t]  # heat of liquid to tray below
            + m.V[c, t - 1] * m.H_V[c, t - 1]  # heat of vapor from tray below
            - m.V[c, t] * m.H_V[c, t]  # heat of vapor to tray above
            for c in m.comps) * 1E-3 == 0

    @tray.Constraint(m.comps)
    def liq_enthalpy_calc(_, c):
        return m.H_L[c, t] == m.liq_enthalpy_expr[t, c]

    @tray.Constraint(m.comps)
    def vap_enthalpy_calc(_, c):
        return m.H_V[c, t] == m.vap_enthalpy_expr[t, c]

    @no_tray.Constraint(m.comps)
    def liq_enthalpy_pass_through(_, c):
        return m.H_L[c, t] == m.H_L[c, t + 1]

    @no_tray.Constraint(m.comps)
    def vap_enthalpy_pass_through(_, c):
        return m.H_V[c, t] == m.H_V[c, t - 1]


def _build_feed_tray_energy_balance(m):
    t = m.feed_tray

    @m.Constraint()
    def feed_tray_energy_balance(_):
        return (
            sum(m.feed[c] * (
                m.H_L_spec_feed[c] * (1 - m.feed_vap_frac) +
                m.H_V_spec_feed[c] * m.feed_vap_frac)
                for c in m.comps) +
            sum(
                # Heat of liquid from tray above
                m.L[c, t + 1] * m.H_L[c, t + 1]
                # heat of liquid to tray below
                - m.L[c, t] * m.H_L[c, t]
                # heat of vapor from tray below
                + m.V[c, t - 1] * m.H_V[c, t - 1]
                # heat of vapor to tray above
                - m.V[c, t] * m.H_V[c, t]
                for c in m.comps)) * 1E-3 == 0

    @m.Constraint(m.comps)
    def feed_tray_liq_enthalpy_calc(_, c):
        return m.H_L[c, t] == m.liq_enthalpy_expr[t, c]

    @m.Constraint(m.comps)
    def feed_tray_vap_enthalpy_calc(_, c):
        return m.H_V[c, t] == m.vap_enthalpy_expr[t, c]

    @m.Expression(m.comps)
    def feed_liq_enthalpy_expr(_, c):
        k = m.liq_Cp_const[c]
        return (
            k['A'] * (m.T_feed - m.T_ref) +
            k['B'] * (m.T_feed ** 2 - m.T_ref ** 2) / 2 +
            k['C'] * (m.T_feed ** 3 - m.T_ref ** 3) / 3 +
            k['D'] * (m.T_feed ** 4 - m.T_ref ** 4) / 4 +
            k['E'] * (m.T_feed ** 5 - m.T_ref ** 5) / 5) * 1E-6

    @m.Constraint(m.comps)
    def feed_liq_enthalpy_calc(_, c):
        return m.H_L_spec_feed[c] == m.feed_liq_enthalpy_expr[c]

    @m.Expression(m.comps)
    def feed_vap_enthalpy_expr(_, c):
        k = m.vap_Cp_const[c]
        return (
            m.dH_vap[c] +
            k['A'] * (m.T_feed - m.T_ref) +
            k['B'] * (m.T_feed ** 2 - m.T_ref ** 2) / 2 +
            k['C'] * (m.T_feed ** 3 - m.T_ref ** 3) / 3 +
            k['D'] * (m.T_feed ** 4 - m.T_ref ** 4) / 4 +
            k['E'] * (m.T_feed ** 5 - m.T_ref ** 5) / 5) * 1E-3

    @m.Constraint(m.comps)
    def feed_vap_enthalpy_calc(_, c):
        return m.H_V_spec_feed[c] == m.feed_vap_enthalpy_expr[c]


def _build_condenser_energy_balance(m):
    t = m.condens_tray

    @m.partial_cond.Constraint()
    def partial_condenser_energy_balance(_):
        return -m.Qc + sum(
            - m.D[c] * m.H_L[c, t]  # heat of liquid distillate
            - m.L[c, t] * m.H_L[c, t]  # heat of liquid to tray below
            + m.V[c, t - 1] * m.H_V[c, t - 1]  # heat of vapor from tray below
            - m.V[c, t] * m.H_V[c, t]  # heat of vapor from partial condenser
            for c in m.comps) * 1E-3 == 0

    @m.total_cond.Constraint()
    def total_condenser_energy_balance(_):
        return -m.Qc + sum(
            - m.D[c] * m.H_L[c, t]  # heat of liquid distillate
            - m.L[c, t] * m.H_L[c, t]  # heat of liquid to tray below
            + m.V[c, t - 1] * m.H_V[c, t - 1]  # heat of vapor from tray below
            for c in m.comps) * 1E-3 == 0

    @m.Constraint(m.comps)
    def condenser_liq_enthalpy_calc(_, c):
        return m.H_L[c, t] == m.liq_enthalpy_expr[t, c]

    @m.partial_cond.Constraint(m.comps)
    def vap_enthalpy_calc(_, c):
        return m.H_V[c, t] == m.vap_enthalpy_expr[t, c]


def _build_reboiler_energy_balance(m):
    t = m.reboil_tray

    @m.Constraint()
    def reboiler_energy_balance(_):
        return m.Qb + sum(
            m.L[c, t + 1] * m.H_L[c, t + 1]  # Heat of liquid from tray above
            - m.B[c] * m.H_L[c, t]  # heat of liquid bottoms if reboiler
            - m.V[c, t] * m.H_V[c, t]  # heat of vapor to tray above
            for c in m.comps) * 1E-3 == 0

    @m.Constraint(m.comps)
    def reboiler_liq_enthalpy_calc(_, c):
        return m.H_L[c, t] == m.liq_enthalpy_expr[t, c]

    @m.Constraint(m.comps)
    def reboiler_vap_enthalpy_calc(_, c):
        return m.H_V[c, t] == m.vap_enthalpy_expr[t, c]



from .batch_processing import build_model as build_batch_processing_model
from .disease_model import build_model as build_disease_model
from .jobshop import build_small_concrete as build_jobshop_model
from .med_term_purchasing import build_concrete as build_med_term_purchasing_model

__all__ = [
    'build_batch_processing_model',
    'build_disease_model',
    'build_jobshop_model',
    'build_med_term_purchasing_model'
]



from pyomo.environ import Suffix

from .column import build_column


def build_model():
    m = build_column(min_trays=8, max_trays=17, xD=0.95, xB=0.95)
    # Fix feed conditions
    m.feed['benzene'].fix(50)
    m.feed['toluene'].fix(50)
    m.T_feed.fix(368)
    m.feed_vap_frac.fix(0.40395)
    # Initial values of reflux and reboil ratios
    m.reflux_ratio.set_value(1.4)
    m.reboil_ratio.set_value(1.3)
    # Fix to be total condenser
    m.partial_cond.deactivate()
    m.total_cond.indicator_var.fix(1)
    # Give initial values of the tray existence/absence
    for t in m.conditional_trays:
        m.tray[t].indicator_var.set_value(1)
        m.no_tray[t].indicator_var.set_value(0)

    m.BigM = Suffix(direction=Suffix.LOCAL)
    m.BigM[None] = 100


__all__ = ['build_model']



# coding: utf-8

# # [Pyomo.GDP](./index.ipynb) Logical Expression System Demo - IR Spectroscopy Parameter Estimation
#
# This is a reproduction of the IR spectroscopy parameter estimation problem found in:
#
# > Vecchietti A. & Grossmann I. E.
# > LOGMIP: A disjunctive 0-1 non-linear optimizer for process system models,
# > *Comp. & Chem Eng.* 23, p. 555-565, 1999.
#
# This code relies on the logic-v1 branch at https://github.com/qtothec/pyomo/tree/logic-v1

# Optimal value: 12.0893

from pyomo.environ import *
from pyomo.gdp import *
from pyomo.core.expr.logical_expr import *
from pyomo.core.plugins.transform.logical_to_linear import update_boolean_vars_from_binary
from six import StringIO
import pandas as pd


def build_model():
    spectroscopic_data = StringIO("""
             1      2       3       4       5       6       7       8
    1     0.0003  0.0764  0.0318  0.0007  0.0534  0.0773  0.0536  0.0320
    2     0.0007  0.0003  0.0004  0.0009  0.0005  0.0009  0.0005  0.0003
    3     0.0066  0.0789  0.0275  0.0043  0.0704  0.0683  0.0842  0.0309
    4     0.0044  0.0186  0.0180  0.0179  0.0351  0.0024  0.0108  0.0052
    5     0.0208  0.0605  0.0601  0.0604  0.0981  0.0025  0.0394  0.0221
    6     0.0518  0.1656  0.1491  0.1385  0.2389  0.0248  0.1122  0.0633
    7     0.0036  0.0035  0.0032  0.0051  0.0015  0.0094  0.0015  0.0024
    8     0.0507  0.0361  0.0433  0.0635  0.0048  0.0891  0.0213  0.0310
    9     0.0905  0.0600  0.0754  0.1098  0.0038  0.1443  0.0420  0.0574
    10    0.0016  0.0209  0.0063  0.0010  0.0132  0.0203  0.0139  0.0057
    """)
    # Note: this could come from an external data file
    spectroscopic_data_table = pd.read_csv(spectroscopic_data, delimiter=r'\s+')
    flat_spectro_data = spectroscopic_data_table.stack()
    spectro_data_dict = {(k[0], int(k[1])): v for k, v in flat_spectro_data.to_dict().items()}  # column labels to integer

    c_data = StringIO("""
            1       2       3       4       5       6       7       8
    1       502     204     353     702     0       1016    104     204
    2        97     351     351     351     700     0       201      97
    3        0      22      8       0       14      22      14       8 
    """)
    c_data_table = pd.read_csv(c_data, delimiter=r'\s+')
    c_data_dict = {(k[0], int(k[1])): v for k, v in c_data_table.stack().to_dict().items()}

    # Covariance matrix
    r_data = StringIO("""
            1       2       3
    1       1       0       0
    2       0       1       0
    3       0       0       1
    """)
    r_data_table = pd.read_csv(r_data, delimiter=r'\s+')
    r_data_dict = {(k[0], int(k[1])): v for k, v in r_data_table.stack().to_dict().items()}

    m = ConcreteModel(name="IR spectroscopy parameter estimation")
    m.wave_number = RangeSet(10)
    m.spectra_data = RangeSet(8)
    m.compounds = RangeSet(3)

    m.A = Param(m.wave_number, m.spectra_data, initialize=spectro_data_dict)
    m.C = Param(m.compounds, m.spectra_data, initialize=c_data_dict)
    m.R = Param(m.compounds, m.compounds, initialize=r_data_dict)

    m.val = Var(m.spectra_data)
    m.ent = Var(m.compounds, m.wave_number, bounds=(0, 1))
    m.Y = BooleanVar(m.compounds, m.wave_number)
    m.P = Var(m.compounds, m.wave_number, bounds=(0, 1000))

    @m.Disjunction(m.compounds, m.wave_number)
    def d(m, k, i):
        return [
            [m.P[k, i] <= 1000, m.P[k, i] >= 0, m.ent[k, i] == 1],
            [m.P[k, i] == 0, m.ent[k, i] == 0]
        ]

    for k, i in m.compounds * m.wave_number:
        m.Y[k, i].set_binary_var(m.d[k, i].disjuncts[0].indicator_var)

    @m.Constraint(m.spectra_data)
    def eq1(m, j):
        return m.val[j] == sum(
            sum((m.C[kk, j] / 100 - sum(m.P[kk, i] * m.A[i, j] for i in m.wave_number))
                * m.R[kk, k]
                for kk in m.compounds)
            * (m.C[k, j] / 100 - sum(m.P[k, i] * m.A[i, j] for i in m.wave_number))
            for k in m.compounds
        )

    m.profit = Objective(
        expr=sum(m.val[j] for j in m.spectra_data) + 2 * sum(m.ent[k, i] for k in m.compounds for i in m.wave_number))

    return m


if __name__ == "__main__":
    m = build_model()
    TransformationFactory('core.logical_to_linear').apply_to(m)
    # res = SolverFactory('gdpopt').solve(m, tee=False, nlp_solver='gams')
    TransformationFactory('gdp.bigm').apply_to(m)
    SolverFactory('gams').solve(m, tee=True, solver='baron')
    update_boolean_vars_from_binary(m)
    m.profit.display()
    m.Y.display()
    m.P.display()





"""Heat integration case study.

This is example 1 of the Yee & Grossmann, 1990 paper "Simultaneous optimization
models for heat integration--II".
DOI: 10.1016/0098-1354(90)85010-8

This is a modification to support the incorporation of standardized exchanger
modules.

This is a further modification to support discretization of area simplifying
the nonlinear expressions, specialized to the case of allowing only a single
exchanger module type (size).

"""
from __future__ import division

from pyomo.environ import Binary, Constraint, RangeSet, Var
from pyomo.gdp import Disjunction

from . import common


def build_single_module(cafaro_approx, num_stages):
    return build_model(cafaro_approx, num_stages)


def build_model(use_cafaro_approximation, num_stages):
    """Build the model."""
    m = common.build_model(use_cafaro_approximation, num_stages)

    # list of tuples (num_modules, module_size)
    configurations_list = []
    for size in m.module_sizes:
        configs = [(i + 1, size) for i in range(m.max_num_modules[size])]
        configurations_list += configs

    # Map of config indx: (# modules, module size)
    m.configurations_map = {
        (k + 1): v for k, v in enumerate(configurations_list)}

    m.module_index_set = RangeSet(len(configurations_list))

    m.module_config_active = Var(
        m.valid_matches, m.stages, m.module_index_set,
        doc="Binary for if which module configuration is active for a match.",
        domain=Binary, initialize=0)

    @m.Param(m.module_index_set, doc="Area of each configuration")
    def module_area(m, indx):
        num_modules, size = m.configurations_map[indx]
        return num_modules * size

    @m.Param(m.valid_matches, m.module_index_set,
             doc="Area cost for each modular configuration.")
    def modular_size_cost(m, hot, cold, indx):
        num_modules, size = m.configurations_map[indx]
        return num_modules * m.module_area_cost[hot, cold, size]

    @m.Param(m.valid_matches, m.module_index_set,
             doc="Fixed cost for each modular exchanger size.")
    def modular_fixed_cost(m, hot, cold, indx):
        num_modules, size = m.configurations_map[indx]
        return num_modules * m.module_fixed_unit_cost

    m.LMTD_discretize = Var(
        m.hot_streams, m.cold_streams, m.stages, m.module_index_set,
        doc="Discretized log mean temperature difference",
        bounds=(0, 500), initialize=0
    )

    for hot, cold, stg in m.valid_matches * m.stages:
        disj = m.exchanger_exists[hot, cold, stg]
        disj.choose_one_config = Constraint(
            expr=sum(
                m.module_config_active[hot, cold, stg, indx]
                for indx in m.module_index_set) == 1
        )

        disj.exchanger_area_cost = Constraint(
            expr=m.exchanger_area_cost[stg, hot, cold] * 1E-3 ==
            sum(m.modular_size_cost[hot, cold, indx] * 1E-3 *
                m.module_config_active[hot, cold, stg, indx]
                for indx in m.module_index_set)
        )
        disj.exchanger_fixed_cost = Constraint(
            expr=m.exchanger_fixed_cost[stg, hot, cold] ==
            sum(m.modular_fixed_cost[hot, cold, indx] * 1E-3 *
                m.module_config_active[hot, cold, stg, indx]
                for indx in m.module_index_set))

        disj.discretize_area = Constraint(
            expr=m.exchanger_area[stg, hot, cold] == sum(
                m.module_area[indx] *
                m.module_config_active[hot, cold, stg, indx]
                for indx in m.module_index_set)
        )

        disj.discretized_LMTD = Constraint(
            expr=m.LMTD[hot, cold, stg] == sum(
                m.LMTD_discretize[hot, cold, stg, indx]
                for indx in m.module_index_set
            )
        )

        @disj.Constraint(m.module_index_set)
        def discretized_LMTD_LB(disj, indx):
            return (
                m.LMTD[hot, cold, stg].lb
                * m.module_config_active[hot, cold, stg, indx]
            ) <= m.LMTD_discretize[hot, cold, stg, indx]

        @disj.Constraint(m.module_index_set)
        def discretized_LMTD_UB(disj, indx):
            return m.LMTD_discretize[hot, cold, stg, indx] <= (
                m.LMTD[hot, cold, stg].ub
                * m.module_config_active[hot, cold, stg, indx]
            )

        disj.exchanger_required_area = Constraint(
            expr=m.U[hot, cold] * sum(
                m.module_area[indx] * m.LMTD_discretize[hot, cold, stg, indx]
                for indx in m.module_index_set) >=
            m.heat_exchanged[hot, cold, stg])

    @m.Disjunct(m.module_sizes)
    def module_type(disj, size):
        """Disjunct for selection of one module type."""
        @disj.Constraint(m.valid_matches, m.stages, m.module_index_set)
        def no_other_module_types(_, hot, cold, stg, indx):
            # num_modules, size = configurations_map[indx]
            if m.configurations_map[indx][1] != size:
                return m.module_config_active[hot, cold, stg, indx] == 0
            else:
                return Constraint.NoConstraint
        # disj.no_other_module_types = Constraint(
        #     expr=sum(
        #         m.module_config_active[hot, cold, stg, indx]
        #         for indx in m.module_index_set
        #         if m.configurations_map[indx][1] != size) == 0
        # )
    m.select_one_module_type = Disjunction(
        expr=[m.module_type[area] for area in m.module_sizes])

    return m



from functools import partial

from .conventional import build_conventional as _conv
from .modular_discrete import build_modular_option as _disc_opt, build_require_modular as _disc_mod
from .modular_discrete_single_module import build_single_module as _disc_sing
from .modular_integer import (
    build_modular_option as _int_opt, build_require_modular as _int_mod,
    build_single_module as _int_sing, )

# These are the functions that we want to expose as public
build_conventional = partial(_conv, cafaro_approx=True, num_stages=4)
build_integer_single_module = partial(_int_sing, cafaro_approx=True, num_stages=4)
build_integer_require_modular = partial(_int_mod, cafaro_approx=True, num_stages=4)
build_integer_modular_option = partial(_int_opt, cafaro_approx=True, num_stages=4)
build_discrete_single_module = partial(_disc_sing, cafaro_approx=True, num_stages=4)
build_discrete_require_modular = partial(_disc_mod, cafaro_approx=True, num_stages=4)
build_discrete_modular_option = partial(_disc_opt, cafaro_approx=True, num_stages=4)

__all__ = ['build_conventional', 'build_integer_single_module', 'build_integer_require_modular',
           'build_integer_modular_option', 'build_discrete_single_module', 'build_discrete_require_modular',
           'build_discrete_modular_option']
